# -*- coding: utf-8 -*-
# ------------------------------------------------------------------
# Filename: waveform.py
#  Purpose: Waveform plotting for obspy.Stream objects
#   Author: Lion Krischer
#    Email: krischer@geophysik.uni-muenchen.de
#
# Copyright (C) 2008-2012 Lion Krischer
# --------------------------------------------------------------------
"""
Waveform plotting for obspy.Stream objects.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA @UnusedWildImport
from future.utils import native_str

import io
import warnings
import functools
from copy import copy
from datetime import datetime

import numpy as np
import matplotlib.lines as mlines
import matplotlib.patches as patches
from matplotlib.cm import get_cmap
from matplotlib.dates import date2num
from matplotlib.path import Path
from matplotlib.ticker import MaxNLocator, ScalarFormatter
import scipy.signal as signal

from obspy import Stream, Trace, UTCDateTime
from obspy.core.util import create_empty_data_chunk, MATPLOTLIB_VERSION
from obspy.geodetics import FlinnEngdahl, kilometer2degrees, locations2degrees
from obspy.imaging.util import (_set_xaxis_obspy_dates, _id_key, _timestring)


MINMAX_ZOOMLEVEL_WARNING_TEXT = "Warning: Zooming into MinMax Plot!"
SECONDS_PER_DAY = 3600.0 * 24.0
DATELOCATOR_WARNING_MSG = (
    "AutoDateLocator was unable to pick an appropriate interval for this date "
    "range. It may be necessary to add an interval value to the "
    "AutoDateLocator's intervald dictionary.")


class WaveformPlotting(object):
    """
    Class that provides several solutions for plotting large and small waveform
    data sets.

    .. warning::

        This class should NOT be used directly, instead use the
        :meth:`~obspy.core.stream.Stream.plot` method of the
        ObsPy :class:`~obspy.core.stream.Stream` or
        :class:`~obspy.core.trace.Trace` objects.

    It uses matplotlib to plot the waveforms.
    """

    def __init__(self, **kwargs):
        """
        Checks some variables and maps the kwargs to class variables.
        """
        self.kwargs = kwargs
        self.stream = kwargs.get('stream')
        # Check if it is a Stream or a Trace object.
        if isinstance(self.stream, Trace):
            self.stream = Stream([self.stream])
        elif not isinstance(self.stream, Stream):
            msg = 'Plotting is only supported for Stream or Trace objects.'
            raise TypeError(msg)
        # Stream object should contain at least one Trace
        if len(self.stream) < 1:
            msg = "Empty stream object"
            raise IndexError(msg)
        self.stream = self.stream.copy()
        # Type of the plot.
        self.type = kwargs.get('type', 'normal')
        # Start and end times of the plots.
        self.starttime = kwargs.get('starttime', None)
        self.endtime = kwargs.get('endtime', None)
        self.fig_obj = kwargs.get('fig', None)
        # If no times are given take the min/max values from the stream object.
        if not self.starttime:
            self.starttime = min([trace.stats.starttime for trace in
                                  self.stream])
        if not self.endtime:
            self.endtime = max([trace.stats.endtime for trace in self.stream])
        self.stream.trim(self.starttime, self.endtime)
        # Assigning values for type 'section'
        self.sect_offset_min = kwargs.get('offset_min', None)
        self.sect_offset_max = kwargs.get('offset_max', None)
        self.sect_dist_degree = kwargs.get('dist_degree', False)
        # TODO Event data from class Event()
        self.ev_coord = kwargs.get('ev_coord', None)
        self.alpha = kwargs.get('alpha', 0.5)
        self.sect_plot_dx = kwargs.get('plot_dx', None)
        if self.sect_plot_dx is not None and not self.sect_dist_degree:
            self.sect_plot_dx /= 1e3
        self.sect_timedown = kwargs.get('time_down', False)
        self.sect_recordstart = kwargs.get('recordstart', None)
        self.sect_recordlength = kwargs.get('recordlength', None)
        self.sect_norm_method = kwargs.get('norm_method', 'trace')
        self.sect_user_scale = kwargs.get('scale', 1.0)
        self.sect_vred = kwargs.get('vred', None)
        if self.sect_vred and self.sect_dist_degree:
            self.sect_vred = kilometer2degrees(self.sect_vred / 1e3)
        self.sect_orientation = kwargs.get('orientation', 'vertical')
        if self.type == 'relative':
            self.reftime = kwargs.get('reftime', self.starttime)
        elif self.type == 'section':
            self.sect_reftime = kwargs.get('reftime', None)
        # Whether to use straight plotting or the fast minmax method. If not
        # set explicitly by the user "full" method will be used by default and
        # "fast" method will be used above some threshold of data points to
        # plot.
        self.plotting_method = kwargs.get('method', None)
        # Below that value the data points will be plotted normally. Above it
        # the data will be plotted using a different approach (details see
        # below). Can be overwritten by the above self.plotting_method kwarg.
        if self.type == 'section':
            # section may consists of hundreds of seismograms
            self.max_npts = 10000
        else:
            self.max_npts = 400000
        # If automerge is enabled, merge traces with the same id for the plot.
        self.automerge = kwargs.get('automerge', True)
        # If equal_scale is enabled, all plots are equally scaled.
        self.equal_scale = kwargs.get('equal_scale', True)
        # Set default values.
        # The default value for the size is determined dynamically because
        # there might be more than one channel to plot.
        self.size = kwargs.get('size', None)
        # Values that will be used to calculate the size of the plot.
        self.default_width = 800
        self.default_height_per_channel = 250
        if not self.size:
            self.width = 800
            # Check the kind of plot.
            if self.type == 'dayplot':
                self.height = 600
            elif self.type == 'section':
                self.width = 1000
                self.height = 600
            else:
                # One plot for each trace.
                if self.automerge:
                    count = self.__get_mergable_ids()
                    count = len(count)
                else:
                    count = len(self.stream)
                self.height = count * 250
        else:
            self.width, self.height = self.size
        # Interval length in minutes for dayplot.
        self.interval = 60 * kwargs.get('interval', 15)
        # Scaling.
        self.vertical_scaling_range = kwargs.get('vertical_scaling_range',
                                                 None)
        # Dots per inch of the plot. Might be useful for printing plots.
        self.dpi = kwargs.get('dpi', 100)
        # Color of the graph.
        if self.type == 'dayplot':
            self.color = kwargs.get('color', ('#B2000F', '#004C12', '#847200',
                                              '#0E01FF'))
            if isinstance(self.color, (str, native_str)):
                self.color = (self.color,)
            self.number_of_ticks = kwargs.get('number_of_ticks', None)
        else:
            self.color = kwargs.get('color', 'k')
            self.number_of_ticks = kwargs.get('number_of_ticks', 4)
        # Background, face and grid color.
        self.background_color = kwargs.get('bgcolor', 'w')
        self.face_color = kwargs.get('face_color', 'w')
        self.grid_color = kwargs.get('grid_color', 'black')
        self.grid_linewidth = kwargs.get('grid_linewidth', 0.5)
        self.grid_linestyle = kwargs.get('grid_linestyle', ':')
        # Transparency. Overwrites background and facecolor settings.
        self.transparent = kwargs.get('transparent', False)
        if self.transparent:
            self.background_color = None
        # Ticks.
        if self.type == 'relative':
            self.tick_format = kwargs.get('tick_format', '%.2f')
        else:
            self.tick_format = kwargs.get('tick_format', '%H:%M:%S')
        self.tick_rotation = kwargs.get('tick_rotation', 0)
        # Whether or not to save a file.
        self.outfile = kwargs.get('outfile')
        self.handle = kwargs.get('handle')
        # File format of the resulting file. Usually defaults to PNG but might
        # be dependent on your matplotlib backend.
        self.format = kwargs.get('format')
        self.show = kwargs.get('show', True)
        self.draw = kwargs.get('draw', True)
        self.block = kwargs.get('block', True)
        # plot parameters options
        self.x_labels_size = kwargs.get('x_labels_size', 8)
        self.y_labels_size = kwargs.get('y_labels_size', 8)
        self.title_size = kwargs.get('title_size', 10)
        self.linewidth = kwargs.get('linewidth', 1)
        self.linestyle = kwargs.get('linestyle', '-')
        self.subplots_adjust_left = kwargs.get('subplots_adjust_left', 0.12)
        self.subplots_adjust_right = kwargs.get('subplots_adjust_right', 0.88)
        self.subplots_adjust_top = kwargs.get('subplots_adjust_top', 0.95)
        self.subplots_adjust_bottom = kwargs.get('subplots_adjust_bottom', 0.1)
        self.right_vertical_labels = kwargs.get('right_vertical_labels', False)
        self.one_tick_per_line = kwargs.get('one_tick_per_line', False)
        self.show_y_UTC_label = kwargs.get('show_y_UTC_label', True)
        self.title = kwargs.get('title', self.stream[0].id)
        self.fillcolor_pos, self.fillcolor_neg = \
            kwargs.get('fillcolors', (None, None))

    def __del__(self):
        """
        Destructor closes the figure instance if it has been created by the
        class.
        """
        import matplotlib.pyplot as plt
        if self.kwargs.get('fig', None) is None and \
                not self.kwargs.get('handle'):
            plt.close()

    def __get_merge_id(self, tr):
        tr_id = tr.id
        # don't merge normal traces with previews
        try:
            if tr.stats.preview:
                tr_id += 'preview'
        except AttributeError:
            pass
        # don't merge traces with different processing steps
        try:
            if tr.stats.processing:
                tr_id += str(tr.stats.processing)
        except AttributeError:
            pass
        return tr_id

    def __get_mergable_ids(self):
        ids = set()
        for tr in self.stream:
            ids.add(self.__get_merge_id(tr))
        return sorted(ids, key=_id_key)

    def plot_waveform(self, *args, **kwargs):
        """
        Creates a graph of any given ObsPy Stream object. It either saves the
        image directly to the file system or returns a binary image string.

        For all color values you can use legit HTML names, HTML hex strings
        (e.g. '#eeefff') or you can pass an RGB tuple, where each of R, G, and
        B are in the range [0, 1]. You can also use single letters for basic
        built-in colors ('b' = blue, 'g' = green, 'r' = red, 'c' = cyan,
        'm' = magenta, 'y' = yellow, 'k' = black, 'w' = white) and gray shades
        can be given as a string encoding a float in the 0-1 range.
        """
        import matplotlib.pyplot as plt
        # Setup the figure if not passed explicitly.
        if not self.fig_obj:
            self.__setup_figure()
        else:
            self.fig = self.fig_obj
        # Determine kind of plot and do the actual plotting.
        if self.type == 'dayplot':
            self.plot_day(*args, **kwargs)
        elif self.type == 'section':
            self.plot_section(*args, **kwargs)
        else:
            self.plot(*args, **kwargs)
        # Adjust the subplot so there is always a fixed margin on every side
        if self.type != 'dayplot':
            fract_y = 60.0 / self.height
            fract_y2 = 40.0 / self.height
            fract_x = 80.0 / self.width
            self.fig.subplots_adjust(top=1.0 - fract_y, bottom=fract_y2,
                                     left=fract_x, right=1.0 - fract_x / 2)
        if self.type == 'section':
            self.fig.subplots_adjust(bottom=0.12)
        with warnings.catch_warnings(record=True):
            warnings.filterwarnings("ignore", DATELOCATOR_WARNING_MSG,
                                    UserWarning, "matplotlib.dates")
            if self.draw:
                self.fig.canvas.draw()
            # The following just serves as a unified way of saving and
            # displaying the plots.
            if not self.transparent:
                extra_args = {'dpi': self.dpi,
                              'facecolor': self.face_color,
                              'edgecolor': self.face_color}
            else:
                extra_args = {'dpi': self.dpi,
                              'transparent': self.transparent,
                              'facecolor': 'k'}
            if self.outfile:
                # If format is set use it.
                if self.format:
                    self.fig.savefig(self.outfile, format=self.format,
                                     **extra_args)
                # Otherwise use format from self.outfile or default to PNG.
                else:
                    self.fig.savefig(self.outfile, **extra_args)
            else:
                # Return a binary image string if not self.outfile but
                # self.format.
                if self.format:
                    imgdata = io.BytesIO()
                    self.fig.savefig(imgdata, format=self.format,
                                     **extra_args)
                    imgdata.seek(0)
                    return imgdata.read()
                elif self.handle:
                    return self.fig
                else:
                    if not self.fig_obj and self.show:
                        try:
                            plt.show(block=self.block)
                        except Exception:
                            plt.show()
                    # Also return the figure to make it work the jupyter
                    # notebooks.
                    return self.fig

    def plot(self, *args, **kwargs):
        """
        Plot the Traces showing one graph per Trace.

        Plots the whole time series for self.max_npts points and less. For more
        points it plots minmax values.
        """
        stream_new = []
        # Just remove empty traces.
        if not self.automerge:
            for tr in self.stream:
                stream_new.append([])
                if len(tr.data):
                    stream_new[-1].append(tr)
        else:
            # Generate sorted list of traces (no copy)
            # Sort order: id, starttime, endtime
            ids = self.__get_mergable_ids()
            for id in ids:
                stream_new.append([])
                for tr in self.stream:
                    tr_id = self.__get_merge_id(tr)
                    if tr_id == id:
                        # does not copy the elements of the data array
                        tr_ref = copy(tr)
                        if tr_ref.data.size:
                            stream_new[-1].append(tr_ref)
                # delete if empty list
                if not len(stream_new[-1]):
                    stream_new.pop()
                    continue
        # If everything is lost in the process raise an Exception.
        if not len(stream_new):
            raise Exception("Nothing to plot")
        # Create helper variable to track ids and min/max/mean values.
        self.ids = []
        # Loop over each Trace and call the appropriate plotting method.
        self.axis = []
        for _i, tr in enumerate(stream_new):
            # Each trace needs to have the same sampling rate.
            sampling_rates = {_tr.stats.sampling_rate for _tr in tr}
            if len(sampling_rates) > 1:
                msg = "All traces with the same id need to have the same " + \
                      "sampling rate."
                raise Exception(msg)
            sampling_rate = sampling_rates.pop()
            if _i == 0:
                sharex = None
            else:
                sharex = self.axis[0]
            # TODO remove once minimum required matplotlib version reaches 2.0
            # see matplotlib/matplotlib#5501
            if MATPLOTLIB_VERSION < [2, 0]:
                axis_facecolor_kwargs = dict(axisbg=self.background_color)
            else:
                axis_facecolor_kwargs = dict(facecolor=self.background_color)
            ax = self.fig.add_subplot(len(stream_new), 1, _i + 1,
                                      sharex=sharex, **axis_facecolor_kwargs)
            self.axis.append(ax)
            # XXX: Also enable the minmax plotting for previews.
            method_ = self.plotting_method
            if method_ is None:
                if ((self.endtime - self.starttime) * sampling_rate >
                        self.max_npts):
                    method_ = "fast"
                else:
                    method_ = "full"
            method_ = method_.lower()
            if method_ == 'full':
                self.__plot_straight(stream_new[_i], ax, *args, **kwargs)
            elif method_ == 'fast':
                self.__plot_min_max(stream_new[_i], ax, *args, **kwargs)
            else:
                msg = "Invalid plot method: '%s'" % method_
                raise ValueError(msg)
        # Set ticks.
        self.__plot_set_x_ticks()
        self.__plot_set_y_ticks()
        xmin = self._time_to_xvalue(self.starttime)
        xmax = self._time_to_xvalue(self.endtime)
        ax.set_xlim(xmin, xmax)
        self._draw_overlap_axvspan_legend()

    def plot_day(self, *args, **kwargs):
        """
        Extend the seismogram.
        """
        # Merge and trim to pad.
        self.stream.merge()
        if len(self.stream) != 1:
            msg = "All traces need to be of the same id for a dayplot"
            raise ValueError(msg)
        self.stream.trim(self.starttime, self.endtime, pad=True)
        # Get minmax array.
        self.__dayplot_get_min_max_values(self, *args, **kwargs)
        # Normalize array
        self.__dayplot_normalize_values(self, *args, **kwargs)
        # Get timezone information. If none is given, use local time.
        self.time_offset = kwargs.get(
            'time_offset',
            round((UTCDateTime(datetime.now()) - UTCDateTime()) / 3600.0, 2))
        self.timezone = kwargs.get('timezone', 'local time')
        # Try to guess how many steps are needed to advance one full time unit.
        self.repeat = None
        intervals = self.extreme_values.shape[0]
        if self.interval < 60 and 60 % self.interval == 0:
            self.repeat = 60 // self.interval
        elif self.interval < 1800 and 3600 % self.interval == 0:
            self.repeat = 3600 // self.interval
        # Otherwise use a maximum value of 10.
        else:
            if intervals >= 10:
                self.repeat = 10
            else:
                self.repeat = intervals
        # Create axis to plot on.
        if self.background_color:
            # TODO remove once minimum required matplotlib version reaches 2.0
            # see matplotlib/matplotlib#5501
            if MATPLOTLIB_VERSION < [2, 0]:
                axis_facecolor_kwargs = dict(axisbg=self.background_color)
            else:
                axis_facecolor_kwargs = dict(facecolor=self.background_color)
            ax = self.fig.add_subplot(1, 1, 1, **axis_facecolor_kwargs)
        else:
            ax = self.fig.add_subplot(1, 1, 1)
        # Adjust the subplots
        self.fig.subplots_adjust(left=self.subplots_adjust_left,
                                 right=self.subplots_adjust_right,
                                 top=self.subplots_adjust_top,
                                 bottom=self.subplots_adjust_bottom)
        # Create x_value_array.
        x_values = np.repeat(np.arange(self.width), 2)
        intervals = self.extreme_values.shape[0]

        for _i in range(intervals):
            # Create offset array.
            y_values = np.ma.empty(self.width * 2)
            y_values.fill(intervals - (_i + 1))
            # Add min and max values.
            y_values[0::2] += self.extreme_values[_i, :, 0]
            y_values[1::2] += self.extreme_values[_i, :, 1]
            # Plot the values.
            ax.plot(x_values, y_values,
                    color=self.color[_i % len(self.color)],
                    linewidth=self.linewidth, linestyle=self.linestyle)
        # Plot the scale, if required.
        scale_unit = kwargs.get("data_unit", None)
        if scale_unit is not None:
            self._plot_dayplot_scale(unit=scale_unit)
        # Set ranges.
        ax.set_xlim(0, self.width - 1)
        ax.set_ylim(-0.3, intervals + 0.3)
        self.axis = [ax]
        # Set ticks.
        self.__dayplot_set_y_ticks(*args, **kwargs)
        self.__dayplot_set_x_ticks(*args, **kwargs)
        # Choose to show grid but only on the x axis.
        self.fig.axes[0].grid(color=self.grid_color,
                              linestyle=self.grid_linestyle,
                              linewidth=self.grid_linewidth)
        self.fig.axes[0].yaxis.grid(False)
        # Set the title of the plot.
        if self.title is not None:
            self.fig.suptitle(self.title, fontsize=self.title_size)
        # Now try to plot some events.
        events = kwargs.get("events", [])
        # Potentially download some events with the help of obspy.clients.fdsn.
        if "min_magnitude" in events:
            try:
                from obspy.clients.fdsn import Client
                c = Client("EMSC")
                events = c.get_events(starttime=self.starttime,
                                      endtime=self.endtime,
                                      minmagnitude=events["min_magnitude"])
            except Exception as e:
                events = None
                msg = "Could not download the events because of '%s: %s'." % \
                    (e.__class__.__name__, str(e))
                warnings.warn(msg)
        if events:
            for event in events:
                self._plot_event(event)

    def _plot_event(self, event):
        """
        Helper function to plot an event into the dayplot.
        """
        ax = self.fig.axes[0]
        seed_id = self.stream[0].id
        if hasattr(event, "preferred_origin"):
            # Get the time from the preferred origin, alternatively the first
            # origin.
            origin = event.preferred_origin()
            if origin is None:
                if event.origins:
                    origin = event.origins[0]
                else:
                    return
            time = origin.time

            # Do the same for the magnitude.
            mag = event.preferred_magnitude()
            if mag is None:
                if event.magnitudes:
                    mag = event.magnitudes[0]
            if mag is None:
                mag = ""
            else:
                mag = "%.1f %s" % (mag.mag, mag.magnitude_type)

            region = FlinnEngdahl().get_region(origin.longitude,
                                               origin.latitude)
            text = region
            if mag:
                text += ", %s" % mag
        else:
            time = event["time"]
            text = event["text"] if "text" in event else None

        # Nothing to do if the event is not on the plot.
        if time < self.starttime or time > self.endtime:
            return
        # Now find the position of the event in plot coordinates.

        def time2xy(time):
            frac = (time - self.starttime) / (self.endtime - self.starttime)
            int_frac = (self.interval) / (self.endtime - self.starttime)
            event_frac = frac / int_frac
            y_pos = self.extreme_values.shape[0] - int(event_frac) - 0.5
            x_pos = (event_frac - int(event_frac)) * self.width
            return x_pos, y_pos
        x_pos, y_pos = time2xy(time)

        if text:
            # Some logic to get a somewhat sane positioning of the annotation
            # box and the arrow..
            text_offset_x = 0.10 * self.width
            text_offset_y = 1.00
            # Relpos determines the connection of the arrow on the box in
            # relative coordinates.
            relpos = [0.0, 0.5]
            # Arc strength is the amount of bending of the arrow.
            arc_strength = 0.25
            if x_pos < (self.width / 2.0):
                text_offset_x_sign = 1.0
                ha = "left"
                # Arc sign determines the direction of bending.
                arc_sign = "+"
            else:
                text_offset_x_sign = -1.0
                ha = "right"
                relpos[0] = 1.0
                arc_sign = "-"
            if y_pos < (self.extreme_values.shape[0] / 2.0):
                text_offset_y_sign = 1.0
                va = "bottom"
            else:
                text_offset_y_sign = -1.0
                va = "top"
                if arc_sign == "-":
                    arc_sign = "+"
                else:
                    arc_sign = "-"

            # Draw the annotation including box.
            ax.annotate(text,
                        # The position of the event.
                        xy=(x_pos, y_pos),
                        # The position of the text, offset depending on the
                        # previously calculated variables.
                        xytext=(x_pos + text_offset_x_sign * text_offset_x,
                                y_pos + text_offset_y_sign * text_offset_y),
                        # Everything in data coordinates.
                        xycoords="data", textcoords="data",
                        # Set the text alignment.
                        ha=ha, va=va,
                        # Text box style.
                        bbox=dict(boxstyle="round", fc="w", alpha=0.6),
                        # Arrow style
                        arrowprops=dict(
                            arrowstyle="-",
                            connectionstyle="arc3, rad=%s%.1f" % (
                                arc_sign, arc_strength),
                            relpos=relpos, shrinkB=7),
                        zorder=10)
        # Draw the actual point. Use a marker with a star shape.
        ax.plot(x_pos, y_pos, "*", color="yellow",
                markersize=12, linewidth=self.linewidth)

        for pick in getattr(event, 'picks', []):
            # check that network/station/location matches
            if pick.waveform_id.get_seed_string().split(".")[:-1] != \
               seed_id.split(".")[:-1]:
                continue
            x_pos, y_pos = time2xy(pick.time)
            ax.plot(x_pos, y_pos, "|", color="red",
                    markersize=50, markeredgewidth=self.linewidth * 4)

    def _plot_dayplot_scale(self, unit):
        """
        Plots the dayplot scale if requested.
        """
        left = self.width
        right = left + 5
        top = 2
        bottom = top - 1

        very_right = right + (right - left)
        middle = bottom + (top - bottom) / 2.0

        verts = [
            (left, top),
            (right, top),
            (right, bottom),
            (left, bottom),
            (right, middle),
            (very_right, middle)
        ]

        codes = [Path.MOVETO,
                 Path.LINETO,
                 Path.LINETO,
                 Path.LINETO,
                 Path.MOVETO,
                 Path.LINETO
                 ]

        path = Path(verts, codes)
        patch = patches.PathPatch(path, lw=1, facecolor="none")
        patch.set_clip_on(False)
        self.fig.axes[0].add_patch(patch)
        factor = self._normalization_factor
        # Manually determine the number of digits after decimal
        if factor >= 1000:
            fmt_string = "%.0f %s"
        elif factor >= 100:
            fmt_string = "%.1f %s"
        else:
            fmt_string = "%.2f %s"
        self.fig.axes[0].text(
            very_right + 3, middle,
            fmt_string % (self._normalization_factor, unit), ha="left",
            va="center", fontsize="small")

    def __plot_straight(self, trace, ax, *args, **kwargs):  # @UnusedVariable
        """
        Just plots the data samples in the self.stream. Useful for smaller
        datasets up to around 1000000 samples (depending on the machine on
        which it's being run).

        Slow and high memory consumption for large datasets.
        """
        # trace argument seems to actually be a list of traces..
        st = Stream(trace)
        self._draw_overlap_axvspans(st, ax)
        for trace in st:
            # Check if it is a preview file and adjust accordingly.
            # XXX: Will look weird if the preview file is too small.
            if trace.stats.get('preview'):
                # Mask the gaps.
                trace.data = np.ma.masked_array(trace.data)
                trace.data[trace.data == -1] = np.ma.masked
                # Recreate the min_max scene.
                dtype = trace.data.dtype
                old_time_range = trace.stats.endtime - trace.stats.starttime
                data = np.empty(2 * trace.stats.npts, dtype=dtype)
                data[0::2] = trace.data / 2.0
                data[1::2] = -trace.data / 2.0
                trace.data = data
                # The times are not supposed to change.
                trace.stats.delta = (
                    old_time_range / float(trace.stats.npts - 1))
            trace.data = np.require(trace.data, np.float64) * trace.stats.calib
            if self.type == 'relative':
                # use seconds of relative sample times and shift by trace's
                # start time, which was set relative to `reftime`.
                x_values = (
                    trace.times() + (trace.stats.starttime - self.reftime))
            else:
                # convert seconds of relative sample times to days and add
                # start time of trace.
                x_values = ((trace.times() / SECONDS_PER_DAY) +
                            date2num(trace.stats.starttime.datetime))
            ax.plot(x_values, trace.data, color=self.color,
                    linewidth=self.linewidth, linestyle=self.linestyle)
        # Write to self.ids
        trace = st[0]
        if trace.stats.get('preview'):
            tr_id = trace.id + ' [preview]'
        elif hasattr(trace, 'label'):
            tr_id = trace.label
        else:
            tr_id = trace.id
        self.ids.append(tr_id)

    def __plot_min_max(self, trace, ax, *args, **kwargs):  # @UnusedVariable
        """
        Plots the data using a min/max approach that calculated the minimum and
        maximum values of each "pixel" and then plots only these values. Works
        much faster with large data sets.
        """
        self._draw_overlap_axvspans(Stream(trace), ax)
        # Some variables to help calculate the values.
        starttime = self._time_to_xvalue(self.starttime)
        endtime = self._time_to_xvalue(self.endtime)
        # The same trace will always have the same sampling_rate.
        sampling_rate = trace[0].stats.sampling_rate
        # width of x axis in seconds
        x_width = endtime - starttime
        # normal plots have x-axis in days, so convert x_width to seconds
        if self.type != "relative":
            x_width = x_width * SECONDS_PER_DAY
        # number of samples that get represented by one min-max pair
        pixel_length = int(
            np.ceil((x_width * sampling_rate + 1) / self.width))
        # Loop over all the traces. Do not merge them as there are many samples
        # and therefore merging would be slow.
        for _i, tr in enumerate(trace):
            trace_length = len(tr.data)
            pixel_count = int(trace_length // pixel_length)
            remaining_samples = int(trace_length % pixel_length)
            remaining_seconds = remaining_samples / sampling_rate
            if self.type != "relative":
                remaining_seconds /= SECONDS_PER_DAY
            # Reference to new data array which does not copy data but can be
            # reshaped.
            if remaining_samples:
                data = tr.data[:-remaining_samples]
            else:
                data = tr.data
            data = data.reshape(pixel_count, pixel_length)
            min_ = data.min(axis=1) * tr.stats.calib
            max_ = data.max(axis=1) * tr.stats.calib
            # Calculate extreme_values and put them into new array.
            if remaining_samples:
                extreme_values = np.empty((pixel_count + 1, 2), dtype=np.float)
                extreme_values[:-1, 0] = min_
                extreme_values[:-1, 1] = max_
                extreme_values[-1, 0] = \
                    tr.data[-remaining_samples:].min() * tr.stats.calib
                extreme_values[-1, 1] = \
                    tr.data[-remaining_samples:].max() * tr.stats.calib
            else:
                extreme_values = np.empty((pixel_count, 2), dtype=np.float)
                extreme_values[:, 0] = min_
                extreme_values[:, 1] = max_
            # Finally plot the data.
            start = self._time_to_xvalue(tr.stats.starttime)
            end = self._time_to_xvalue(tr.stats.endtime)
            if remaining_samples:
                # the last minmax pair is inconsistent regarding x-spacing
                x_values = np.linspace(start, end - remaining_seconds,
                                       num=extreme_values.shape[0] - 1)
                x_values = np.concatenate([x_values, [end]])
            else:
                x_values = np.linspace(start, end, num=extreme_values.shape[0])
            x_values = np.repeat(x_values, 2)
            y_values = extreme_values.flatten()
            ax.plot(x_values, y_values, color=self.color)
        # remember xlim state and add callback to warn when zooming in
        self._initial_xrange = (self._time_to_xvalue(self.endtime) -
                                self._time_to_xvalue(self.starttime))
        self._minmax_plot_xrange_dangerous = False
        ax.callbacks.connect("xlim_changed", self._warn_on_xaxis_zoom)
        # set label, write to self.ids
        if hasattr(trace[0], 'label'):
            tr_id = trace[0].label
        else:
            tr_id = trace[0].id
        self.ids.append(tr_id)

    def __plot_set_x_ticks(self, *args, **kwargs):  # @UnusedVariable
        """
        Goes through all axes in pyplot and sets time ticks on the x axis.
        """
        import matplotlib.pyplot as plt
        self.fig.subplots_adjust(hspace=0)
        # Loop over all but last axes.
        for ax in self.axis[:-1]:
            plt.setp(ax.get_xticklabels(), visible=False)
        # set bottom most axes:
        ax = self.axis[-1]
        if self.type == "relative":
            locator = MaxNLocator(5)
            ax.xaxis.set_major_locator(locator)
        else:
            _set_xaxis_obspy_dates(ax)
        plt.setp(ax.get_xticklabels(), fontsize='small',
                 rotation=self.tick_rotation)

    def __plot_set_y_ticks(self, *args, **kwargs):  # @UnusedVariable
        """
        """
        import matplotlib.pyplot as plt
        if self.equal_scale:
            ylims = np.vstack([ax.get_ylim() for ax in self.axis])
            yranges = np.diff(ylims).flatten()
            yrange_max = yranges.max()
            yrange_paddings = -yranges + yrange_max
            ylims[:, 0] -= yrange_paddings[:] / 2
            ylims[:, 1] += yrange_paddings[:] / 2
            for ax, ylims_ in zip(self.axis, ylims):
                ax.set_ylim(*ylims_)
        for _i, ax in enumerate(self.axis):
            # Set the title of each plot.
            ax.text(0.02, 0.95, self.ids[_i], transform=ax.transAxes,
                    fontdict=dict(fontsize="small", ha='left', va='top'),
                    bbox=dict(boxstyle="round", fc="w", alpha=0.8))
            plt.setp(ax.get_yticklabels(), fontsize='small')
            ax.yaxis.set_major_locator(MaxNLocator(7, prune="both"))
            ax.yaxis.set_major_formatter(ScalarFormatter())

    def __dayplot_get_min_max_values(self, *args, **kwargs):  # @UnusedVariable
        """
        Takes a Stream object and calculates the min and max values for each
        pixel in the dayplot.

        Writes a three dimensional array. The first axis is the step, i.e
        number of trace, the second is the pixel in that step and the third
        contains the minimum and maximum value of the pixel.
        """
        # Helper variables for easier access.
        trace = self.stream[0]
        trace_length = len(trace.data)

        # Samples per interval.
        spi = int(self.interval * trace.stats.sampling_rate)
        # Check the approximate number of samples per pixel and raise
        # error as fit.
        spp = float(spi) / self.width
        if spp < 1.0:
            msg = """
            Too few samples to use dayplot with the given arguments.
            Adjust your arguments or use a different plotting method.
            """
            msg = " ".join(msg.strip().split())
            raise ValueError(msg)
        # Number of intervals plotted.
        noi = float(trace_length) / spi
        inoi = int(round(noi))
        # Plot an extra interval if at least 2 percent of the last interval
        # will actually contain data. Do it this way to lessen floating point
        # inaccuracies.
        if abs(noi - inoi) > 2E-2:
            noi = inoi + 1
        else:
            noi = inoi

        # Adjust data. Fill with masked values in case it is necessary.
        number_of_samples = noi * spi
        delta = number_of_samples - trace_length
        if delta < 0:
            trace.data = trace.data[:number_of_samples]
        elif delta > 0:
            trace.data = np.ma.concatenate(
                [trace.data, create_empty_data_chunk(delta, trace.data.dtype)])

        # Create array for min/max values. Use masked arrays to handle gaps.
        extreme_values = np.ma.empty((noi, self.width, 2))
        trace.data.shape = (noi, spi)

        ispp = int(spp)
        fspp = spp % 1.0
        if fspp == 0.0:
            delta = None
        else:
            delta = spi - ispp * self.width

        # Loop over each interval to avoid larger errors towards the end.
        for _i in range(noi):
            if delta:
                cur_interval = trace.data[_i][:-delta]
                rest = trace.data[_i][-delta:]
            else:
                cur_interval = trace.data[_i]
            cur_interval.shape = (self.width, ispp)
            extreme_values[_i, :, 0] = cur_interval.min(axis=1)
            extreme_values[_i, :, 1] = cur_interval.max(axis=1)
            # Add the rest.
            if delta:
                extreme_values[_i, -1, 0] = min(extreme_values[_i, -1, 0],
                                                rest.min())
                extreme_values[_i, -1, 1] = max(extreme_values[_i, -1, 0],
                                                rest.max())
        # Set class variable.
        self.extreme_values = extreme_values

    def __dayplot_normalize_values(self, *args, **kwargs):  # @UnusedVariable
        """
        Normalizes all values in the 3 dimensional array, so that the minimum
        value will be 0 and the maximum value will be 1.

        It will also convert all values to floats.
        """
        # Convert to native floats.
        self.extreme_values = self.extreme_values.astype(np.float) * \
            self.stream[0].stats.calib
        # Make sure that the mean value is at 0
        # raises underflow warning / error for numpy 1.9
        # even though mean is 0.09
        # self.extreme_values -= self.extreme_values.mean()
        self.extreme_values -= self.extreme_values.sum() / \
            self.extreme_values.size

        # Scale so that 99.5 % of the data will fit the given range.
        if self.vertical_scaling_range is None:
            percentile_delta = 0.005
            max_values = self.extreme_values[:, :, 1].compressed()
            min_values = self.extreme_values[:, :, 0].compressed()
            # Remove masked values.
            max_values.sort()
            min_values.sort()
            length = len(max_values)
            index = int((1.0 - percentile_delta) * length)
            max_val = max_values[index]
            index = int(percentile_delta * length)
            min_val = min_values[index]
        # Exact fit.
        elif float(self.vertical_scaling_range) == 0.0:
            max_val = self.extreme_values[:, :, 1].max()
            min_val = self.extreme_values[:, :, 0].min()
        # Fit with custom range.
        else:
            max_val = min_val = abs(self.vertical_scaling_range) / 2.0

        # Normalization factor.
        self._normalization_factor = max(abs(max_val), abs(min_val)) * 2

        # Scale from 0 to 1.
        # raises underflow warning / error for numpy 1.9
        # even though normalization_factor is 2.5
        # self.extreme_values = self.extreme_values / \
        #     self._normalization_factor
        self.extreme_values = self.extreme_values * \
            (1. / self._normalization_factor)
        self.extreme_values += 0.5

    def __dayplot_set_x_ticks(self, *args, **kwargs):  # @UnusedVariable
        """
        Sets the xticks for the dayplot.
        """
        localization_dict = kwargs.get('localization_dict', {})
        localization_dict.setdefault('seconds', 'seconds')
        localization_dict.setdefault('minutes', 'minutes')
        localization_dict.setdefault('hours', 'hours')
        localization_dict.setdefault('time in', 'time in')
        max_value = self.width - 1
        # Check whether it is sec/mins/hours and convert to a universal unit.
        if self.interval < 240:
            time_type = localization_dict['seconds']
            time_value = self.interval
        elif self.interval < 24000:
            time_type = localization_dict['minutes']
            time_value = self.interval / 60
        else:
            time_type = localization_dict['hours']
            time_value = self.interval / 3600
        count = None
        # Hardcode some common values. The plus one is intentional. It had
        # hardly any performance impact and enhances readability.
        if self.interval == 15 * 60:
            count = 15 + 1
        elif self.interval == 20 * 60:
            count = 4 + 1
        elif self.interval == 30 * 60:
            count = 6 + 1
        elif self.interval == 60 * 60:
            count = 4 + 1
        elif self.interval == 90 * 60:
            count = 6 + 1
        elif self.interval == 120 * 60:
            count = 4 + 1
        elif self.interval == 180 * 60:
            count = 6 + 1
        elif self.interval == 240 * 60:
            count = 6 + 1
        elif self.interval == 300 * 60:
            count = 6 + 1
        elif self.interval == 360 * 60:
            count = 12 + 1
        elif self.interval == 720 * 60:
            count = 12 + 1
        # Otherwise run some kind of autodetection routine.
        if not count:
            # Up to 15 time units and if it's a full number, show every unit.
            if time_value <= 15 and time_value % 1 == 0:
                count = time_value
            # Otherwise determine whether they are divisible for numbers up to
            # 15. If a number is not divisible just show 10 units.
            else:
                count = 10
                for _i in range(15, 1, -1):
                    if time_value % _i == 0:
                        count = _i
                        break
            # Show at least 5 ticks.
            if count < 5:
                count = 5
        # Everything can be overwritten by user-specified number of ticks.
        if self.number_of_ticks:
            count = self.number_of_ticks
        # Calculate and set ticks.
        ticks = np.linspace(0.0, max_value, count)
        ticklabels = ['%i' % _i for _i in np.linspace(0.0, time_value, count)]
        self.axis[0].set_xticks(ticks)
        self.axis[0].set_xticklabels(ticklabels, rotation=self.tick_rotation,
                                     size=self.x_labels_size)
        self.axis[0].set_xlabel('%s %s' % (localization_dict['time in'],
                                           time_type), size=self.x_labels_size)

    def __dayplot_set_y_ticks(self, *args, **kwargs):  # @UnusedVariable
        """
        Sets the yticks for the dayplot.
        """
        intervals = self.extreme_values.shape[0]
        # Only display all ticks if there are five or less steps or if option
        # is set.
        if intervals <= 5 or self.one_tick_per_line:
            tick_steps = list(range(0, intervals))
            ticks = np.arange(intervals, 0, -1, dtype=np.float)
            ticks -= 0.5
        else:
            tick_steps = list(range(0, intervals, self.repeat))
            ticks = np.arange(intervals, 0, -1 * self.repeat, dtype=np.float)
            ticks -= 0.5

        # Complicated way to calculate the label of
        # the y-axis showing the second time zone.
        sign = '%+i' % self.time_offset
        sign = sign[0]
        label = "UTC (%s = UTC %s %02i:%02i)" % (
            self.timezone.strip(), sign, abs(self.time_offset),
            (self.time_offset % 1 * 60))
        ticklabels = [(self.starttime + _i *
                       self.interval).strftime(self.tick_format)
                      for _i in tick_steps]
        self.axis[0].set_yticks(ticks)
        self.axis[0].set_yticklabels(ticklabels, size=self.y_labels_size)
        # Show time zone label if requested
        if self.show_y_UTC_label:
            self.axis[0].set_ylabel(label)
        if self.right_vertical_labels:
            yrange = self.axis[0].get_ylim()
            self.twin_x = self.axis[0].twinx()
            self.twin_x.set_ylim(yrange)
            self.twin_x.set_yticks(ticks)
            y_ticklabels_twin = [(self.starttime + (_i + 1) *
                                  self.interval).strftime(self.tick_format)
                                 for _i in tick_steps]
            self.twin_x.set_yticklabels(y_ticklabels_twin,
                                        size=self.y_labels_size)

    def plot_section(self, *args, **kwargs):  # @UnusedVariable
        """
        Plots multiple waveforms as a record section on a single plot.
        """
        # Initialise data and plot
        self.__sect_init_traces()
        ax, lines = self.__sect_init_plot()

        # Setting up line properties
        try:
            names_colors = self.sect_color.items()
        except AttributeError:
            for line in lines:
                line.set_alpha(self.alpha)
                line.set_linewidth(self.linewidth)
                line.set_color(self.sect_color)
        else:
            for line, tr in zip(lines, self.stream):
                line.set_alpha(self.alpha)
                line.set_linewidth(self.linewidth)
                color = self.sect_color[getattr(tr.stats, self.color)]
                line.set_color(color)

            legend_lines = []
            legend_labels = []
            for name, color in sorted(names_colors):
                legend_lines.append(
                    mlines.Line2D([], [], color=color, alpha=self.alpha,
                                  linewidth=self.linewidth))
                legend_labels.append(name)
            ax.legend(legend_lines, legend_labels)

        # Setting up plot axes
        if self.sect_orientation == 'vertical':
            if self.sect_offset_min is not None:
                self.set_offset_lim(left=self._offset_min)
            if self.sect_offset_max is not None:
                self.set_offset_lim(right=self._offset_max)
        elif self.sect_orientation == 'horizontal':
            if self.sect_offset_min is not None:
                self.set_offset_lim(bottom=self._offset_min)
            if self.sect_offset_max is not None:
                self.set_offset_lim(top=self._offset_max)

        # Set up offset ticks
        tick_min, tick_max = np.array(self.get_offset_lim())
        if tick_min != 0.0 and self.sect_plot_dx is not None:
            tick_min += self.sect_plot_dx - (tick_min % self.sect_plot_dx)
        # Define tick vector for offset axis
        ticks = None
        if self.sect_plot_dx:
            xmin, xmax = ax.get_xlim()
            ticks = np.concatenate((
                np.arange(-self.sect_plot_dx, xmin, -self.sect_plot_dx)[::-1],
                np.arange(0, xmax, self.sect_plot_dx)))
            if len(ticks) > 100:
                self.fig.clf()
                msg = 'Too many ticks! Try changing plot_dx.'
                raise ValueError(msg)
            self.set_offset_ticks(ticks)
        # Setting up tick labels
        self.set_time_label('Time [s]')
        if not self.sect_dist_degree:
            self.set_offset_label('Offset [km]')
        else:
            self.set_offset_label(u'Offset [°]')
        if ticks is not None:
            self.set_offset_ticklabels(ticks)
        ax.minorticks_on()
        # Limit time axis
        self.set_time_lim([self._time_min, self._time_max])
        if self.sect_recordstart is not None:
            self.set_time_lim(bottom=self.sect_recordstart)
        if self.sect_recordlength is not None:
            self.set_time_lim(
                top=self.sect_recordlength + self.get_time_lim()[0])
        # Invert time axis if requested
        if self.sect_orientation == 'vertical' and self.sect_timedown:
            ax.invert_yaxis()
        # Draw grid on xaxis only
        ax.grid(
            color=self.grid_color,
            linestyle=self.grid_linestyle,
            linewidth=self.grid_linewidth)
        self.offset_axis.grid(False)

    def __sect_init_traces(self):
        """
        Arrange the trace data used for plotting.

        If necessary the data is resampled before
        being collected in a continuous list.
        """
        # Extract distances from st[].stats.distance
        # or from st.[].stats.coordinates.latitude...
        self._tr_offsets = np.empty(len(self.stream))
        if not self.sect_dist_degree:
            # Define offset in km from tr.stats.distance
            try:
                for _i, tr in enumerate(self.stream):
                    self._tr_offsets[_i] = tr.stats.distance
            except Exception:
                msg = 'trace.stats.distance undefined ' +\
                      '(set before plotting [in m], ' +\
                      'or use the ev_coords argument)'
                raise ValueError(msg)
        else:
            # Define offset as degree from epicenter
            try:
                for _i, tr in enumerate(self.stream):
                    self._tr_offsets[_i] = locations2degrees(
                        tr.stats.coordinates.latitude,
                        tr.stats.coordinates.longitude,
                        self.ev_coord[0], self.ev_coord[1])
            except Exception:
                msg = 'Define latitude/longitude in trace.stats.' + \
                    'coordinates and ev_coord. See documentation.'
                raise ValueError(msg)
        # Define minimum and maximum offsets
        if self.sect_offset_min is None:
            self._offset_min = self._tr_offsets.min()
        else:
            self._offset_min = self.sect_offset_min

        if self.sect_offset_max is None:
            self._offset_max = self._tr_offsets.max()
        else:
            self._offset_max = self.sect_offset_max
        # Reduce data to indexes within offset_min/max
        mask = ((self._tr_offsets >= self._offset_min) &
                (self._tr_offsets <= self._offset_max))
        self._tr_offsets = self._tr_offsets[mask]
        self.stream = [tr for m, tr in zip(mask, self.stream) if m]
        # Use km on distance axis, if not degrees
        if not self.sect_dist_degree:
            self._tr_offsets /= 1e3
            self._offset_min /= 1e3
            self._offset_max /= 1e3
        # Number of traces
        self._tr_num = len(self._tr_offsets)
        # Arranging trace data in single list
        self._tr_data = []
        # Maximum counts, npts, starttime and delta of each selected trace
        self._tr_starttimes = []
        self._tr_max_count = np.empty(self._tr_num)
        self._tr_npts = np.empty(self._tr_num)
        self._tr_delta = np.empty(self._tr_num)
        # TODO dynamic DATA_MAXLENGTH according to dpi
        for _i, tr in enumerate(self.stream):
            if len(tr.data) >= self.max_npts:
                tmp_data = signal.resample(tr.data, self.max_npts)
            else:
                tmp_data = tr.data
            # Initialising trace stats
            self._tr_data.append(tmp_data)
            self._tr_starttimes.append(tr.stats.starttime)
            self._tr_max_count[_i] = tmp_data.max()
            self._tr_npts[_i] = tmp_data.size
            self._tr_delta[_i] = (
                tr.stats.endtime -
                tr.stats.starttime) / self._tr_npts[_i]
        # Init time vectors
        self.__sect_init_time()
        # Init trace colors
        self.__sect_init_color()

    def __sect_scale_traces(self):
        """
        The traces have to be scaled to fit between 0-1., each trace
        gets distance-range/num_traces space. adjustable by scale=1.0.
        """
        self._sect_scale = (
            (self._offset_max - self._offset_min) * self.sect_user_scale /
            (self._tr_num * 1.5))

    def __sect_init_time(self):
        """
        Define the time vector for each trace
        """
        reftime = self.sect_reftime or min(self._tr_starttimes)
        self._tr_times = []
        for _tr in range(self._tr_num):
            self._tr_times.append(
                np.arange(self._tr_npts[_tr]) * self._tr_delta[_tr] +
                (self._tr_starttimes[_tr] - reftime))
            if self.sect_vred:
                self._tr_times[-1] -= self._tr_offsets[_tr] / self.sect_vred

        self._time_min = np.concatenate(self._tr_times).min()
        self._time_max = np.concatenate(self._tr_times).max()

    def __sect_init_color(self):
        """
        Define the color of each trace
        """
        if self.color == 'network':
            colors = {_tr.stats.network for _tr in self.stream}
        elif self.color == 'station':
            colors = {_tr.stats.station for _tr in self.stream}
        elif self.color == 'channel':
            colors = {_tr.stats.channel for _tr in self.stream}
        else:
            self.sect_color = self.color
            return

        cmap = get_cmap('Paired', lut=len(colors))
        self.sect_color = {k: cmap(i) for i, k in enumerate(sorted(colors))}

    def __sect_fraction_to_offset(self, fraction):
        """
        Helper function to return fractions from offsets
        """
        return fraction * self._tr_offsets.max()

    def __sect_init_plot(self):
        """
        Function initialises plot all the illustration is done by
        self.plot_section()
        """
        ax = self.fig.gca()
        # Matplotlib 1.5.x does not support interpolation on fill_betweenx
        # Should be integrated by version 2.1
        # (see https://github.com/matplotlib/matplotlib/pull/6560)
        fill_kwargs = {"lw": 0}
        # There's a strange problem with matplotlib 1.4.1 and 1.4.2.
        # It seems to not render the filled PolyCollection
        # objects if linewidth is set to 0 (see http://tests.obspy.org/48219/,
        # #1502). To circumvent this, use a line color with alpha 0 (and a very
        # small linewidth).
        if [1, 4, 1] <= MATPLOTLIB_VERSION < [1, 4, 3]:
            fill_kwargs = {"edgecolor": (0, 0, 0, 0), "lw": 0.01}

        if self.sect_orientation == 'vertical':
            self.fillfun = functools.partial(ax.fill_betweenx, **fill_kwargs)
        else:
            self.fillfun = functools.partial(ax.fill_between, interpolate=True,
                                             **fill_kwargs)
        # Calculate normalizing factor
        self.__sect_normalize_traces()
        # Calculate scaling factor
        self.__sect_scale_traces()
        lines = []
        # ax.plot() preferred over containers
        for _tr in range(self._tr_num):
            # Scale, normalize and shift traces by offset for plotting
            data = ((self._tr_data[_tr] / self._tr_normfac[_tr] *
                     self._sect_scale) +
                    self._tr_offsets[_tr])
            time = self._tr_times[_tr]
            if self.sect_orientation == 'vertical':
                lines += ax.plot(data, time)
            elif self.sect_orientation == 'horizontal':
                lines += ax.plot(time, data)
            else:
                raise NotImplementedError("sect_orientiation '%s' is not "
                                          "valid." % self.sect_orientation)
            if self.fillcolor_pos:
                self.fillfun(time, data, self._tr_offsets[_tr],
                             where=data > self._tr_offsets[_tr],
                             facecolor=self.fillcolor_pos)
            if self.fillcolor_neg:
                self.fillfun(time, data, self._tr_offsets[_tr],
                             where=data < self._tr_offsets[_tr],
                             facecolor=self.fillcolor_neg)

        # Set correct axes orientation
        if self.sect_orientation == 'vertical':
            self.set_offset_lim = ax.set_xlim
            self.get_offset_lim = ax.get_xlim
            self.set_offset_ticks = ax.set_xticks
            self.set_offset_label = ax.set_xlabel
            self.set_offset_ticklabels = ax.set_xticklabels
            self.offset_axis = ax.xaxis
            self.set_time_lim = ax.set_ylim
            self.get_time_lim = ax.get_ylim
            self.set_time_label = ax.set_ylabel
        elif self.sect_orientation == 'horizontal':
            def _set_xlim_from_ylim(*args, **kwargs):
                if 'bottom' in kwargs:
                    kwargs['left'] = kwargs.pop('bottom')
                if 'top' in kwargs:
                    kwargs['right'] = kwargs.pop('top')
                ax.set_xlim(*args, **kwargs)

            self.set_offset_lim = ax.set_ylim
            self.get_offset_lim = ax.get_ylim
            self.set_offset_ticks = ax.set_yticks
            self.set_offset_label = ax.set_ylabel
            self.set_offset_ticklabels = ax.set_yticklabels
            self.offset_axis = ax.yaxis
            self.set_time_lim = _set_xlim_from_ylim
            self.get_time_lim = ax.get_xlim
            self.set_time_label = ax.set_xlabel
        else:
            raise NotImplementedError("sect_orientiation '%s' is not "
                                      "valid." % self.sect_orientation)

        return ax, lines

    def __sect_normalize_traces(self):
        """
        This helper function normalizes the traces
        """
        self._tr_normfac = np.ones(self._tr_num)
        if self.sect_norm_method == 'trace':
            # Normalize against each traces' maximum
            for tr in range(self._tr_num):
                self._tr_normfac[tr] = np.abs(self._tr_data[tr]).max()
        elif self.sect_norm_method == 'stream':
            # Normalize the whole stream
            tr_max_count_glob = np.abs(self._tr_max_count).max()
            self._tr_normfac.fill(tr_max_count_glob)
        else:
            msg = 'Define a normalisation method. Valid normalisations' + \
                'are \'trace\', \'stream\'. See documentation.'
            raise ValueError(msg)

    def __setup_figure(self):
        """
        The design and look of the whole plot to be produced.
        """
        import matplotlib.pyplot as plt
        # Setup figure and axes
        self.fig = plt.figure(num=None, dpi=self.dpi,
                              figsize=(float(self.width) / self.dpi,
                                       float(self.height) / self.dpi))
        # XXX: Figure out why this is needed sometimes.
        # Set size and dpi.
        self.fig.set_dpi(self.dpi)
        self.fig.set_figwidth(float(self.width) / self.dpi)
        self.fig.set_figheight(float(self.height) / self.dpi)

        if hasattr(self.stream, 'label'):
            suptitle = self.stream.label
        elif self.type == 'relative':
            suptitle = ("Time in seconds relative to %s" %
                        _timestring(self.reftime))
        elif self.type == 'dayplot':
            suptitle = '%s %s' % (self.stream[0].id,
                                  self.starttime.strftime('%Y-%m-%d'))
        elif self.type == 'section':
            suptitle = 'Network: %s [%s] - (%i traces / %s)' % \
                (self.stream[-1].stats.network, self.stream[-1].stats.channel,
                 len(self.stream), _timestring(self.starttime))
        else:
            suptitle = '%s  -  %s' % (_timestring(self.starttime),
                                      _timestring(self.endtime))
        # add suptitle
        y = (self.height - 15.0) / self.height
        self.fig.suptitle(suptitle, y=y, fontsize='small',
                          horizontalalignment='center')

    def _warn_on_xaxis_zoom(self, ax):
        """
        Method to be used as a callback on `method=fast`, "minmax"-type plots
        to warn the user when zooming into the plot.
        """
        xlim = ax.get_xlim()
        if xlim[1] - xlim[0] < 0.9 * self._initial_xrange:
            dangerous = True
        else:
            dangerous = False
        if dangerous and not self._minmax_plot_xrange_dangerous:
            self._add_zoomlevel_warning_text()
        elif self._minmax_plot_xrange_dangerous and not dangerous:
            self._remove_zoomlevel_warning_text()
        self._minmax_plot_xrange_dangerous = dangerous

    def _add_zoomlevel_warning_text(self):
        ax = self.fig.axes[0]
        self._minmax_warning_text = ax.text(
            0.95, 0.9, MINMAX_ZOOMLEVEL_WARNING_TEXT, color="r",
            ha="right", va="top", transform=ax.transAxes)

    def _remove_zoomlevel_warning_text(self):
        ax = self.fig.axes[0]
        if self._minmax_warning_text in ax.texts:
            ax.texts.remove(self._minmax_warning_text)
        self._minmax_warning_text = None

    def _draw_overlap_axvspans(self, st, ax):
        for _, _, _, _, start, end, delta, _ in st.get_gaps():
            if delta > 0:
                continue
            start = self._time_to_xvalue(start)
            end = self._time_to_xvalue(end)
            self._overlap_axvspan = \
                ax.axvspan(start, end, color="r", zorder=-10, alpha=0.5)

    def _draw_overlap_axvspan_legend(self):
        if hasattr(self, "_overlap_axvspan"):
            self.fig.axes[-1].legend(
                [self._overlap_axvspan], ["Overlaps"],
                loc="lower right", prop=dict(size="small"))

    def _time_to_xvalue(self, t):
        if self.type == 'relative':
            return t - self.reftime
        else:
            return date2num(t.datetime)


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



# -*- coding: utf-8 -*-
"""
Python Module for (Weighted) Linear Regression.

:authors:
    Thomas Lecocq (thomas.lecocq@seismology.be)

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA

import numpy as np
import scipy.optimize


def linear_regression(xdata, ydata, weights=None, p0=None,
                      intercept_origin=True, **kwargs):
    """
    Use linear least squares to fit a function, f, to data.
    This method is a generalized version of
    :func:`scipy.optimize.minpack.curve_fit`; allowing for Ordinary Least
    Square and Weighted Least Square regressions:

    * OLS through origin: ``linear_regression(xdata, ydata)``
    * OLS with any intercept: ``linear_regression(xdata, ydata,
      intercept_origin=False)``
    * WLS through origin: ``linear_regression(xdata, ydata, weights)``
    * WLS with any intercept: ``linear_regression(xdata, ydata, weights,
      intercept_origin=False)``

    If the expected values of slope (and intercept) are different from 0.0,
    provide the p0 value(s).

    :param xdata: The independent variable where the data is measured.
    :param ydata: The dependent data - nominally f(xdata, ...)
    :param weights: If not None, the uncertainties in the ydata array. These
        are used as weights in the least-squares problem. If ``None``, the
        uncertainties are assumed to be 1. In SciPy vocabulary, our weights are
        1/sigma.
    :param p0: Initial guess for the parameters. If ``None``, then the initial
        values will all be 0 (Different from SciPy where all are 1)
    :param intercept_origin: If ``True``: solves ``y=a*x`` (default);
        if ``False``: solves ``y=a*x+b``.

    Extra keword arguments will be passed to
    :func:`scipy.optimize.minpack.curve_fit`.

    :rtype: tuple
    :returns: (slope, std_slope) if ``intercept_origin`` is ``True``;
        (slope, intercept, std_slope, std_intercept) if ``False``.
    """
    if weights is not None:
        sigma = 1. / weights
    else:
        sigma = None

    if p0 is None:
        if intercept_origin:
            p0 = 0.0
        else:
            p0 = [0.0, 0.0]

    if intercept_origin:
        p, cov = scipy.optimize.curve_fit(lambda x, a: a * x,
                                          xdata, ydata, p0, sigma=sigma,
                                          **kwargs)
        slope = p[0]
        std_slope = np.sqrt(cov[0, 0])
        return slope, std_slope

    else:
        p, cov = scipy.optimize.curve_fit(lambda x, a, b: a * x + b,
                                          xdata, ydata, p0, sigma=sigma,
                                          **kwargs)
        slope, intercept = p
        std_slope = np.sqrt(cov[0, 0])
        std_intercept = np.sqrt(cov[1, 1])
        return slope, intercept, std_slope, std_intercept


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



# -*- coding: utf-8 -*-
"""
The obspy.io.segy test suite.
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA

import io
import os
import unittest
import warnings

import numpy as np

import obspy
from obspy.core.compatibility import from_buffer, mock
from obspy.core.util import NamedTemporaryFile, AttribDict
from obspy.io.segy.header import (DATA_SAMPLE_FORMAT_PACK_FUNCTIONS,
                                  DATA_SAMPLE_FORMAT_UNPACK_FUNCTIONS)
from obspy.io.segy.segy import (SEGYBinaryFileHeader, SEGYFile,
                                SEGYTraceHeader, _read_segy, iread_segy,
                                SEGYInvalidTextualHeaderWarning)
from obspy.io.segy.tests.header import DTYPES, FILES

from . import _patch_header


class SEGYTestCase(unittest.TestCase):
    """
    Test cases for SEG Y reading and writing..
    """
    def setUp(self):
        # directory where the test files are located
        self.dir = os.path.dirname(__file__)
        self.path = os.path.join(self.dir, 'data')
        # All the files and information about them. These files will be used in
        # most tests. data_sample_enc is the encoding of the data value and
        # sample_size the size in bytes of these samples.
        self.files = FILES
        self.dtypes = DTYPES

    def test_unpack_segy_data(self):
        """
        Tests the unpacking of various SEG Y files.
        """
        for file, attribs in self.files.items():
            data_format = attribs['data_sample_enc']
            endian = attribs['endian']
            count = attribs['sample_count']
            file = os.path.join(self.path, file)
            # Use the with statement to make sure the file closes.
            with open(file, 'rb') as f:
                # Jump to the beginning of the data.
                f.seek(3840)
                # Unpack the data.
                data = DATA_SAMPLE_FORMAT_UNPACK_FUNCTIONS[data_format](
                    f, count, endian)
            # Check the dtype of the data.
            self.assertEqual(data.dtype, self.dtypes[data_format])
            # Proven data values, read with Madagascar.
            correct_data = np.load(file + '.npy').ravel()
            # Compare both.
            np.testing.assert_array_equal(correct_data, data)

    def test_pack_segy_data(self):
        """
        Tests the packing of various SEG Y files.
        """
        # Loop over all files.
        for file, attribs in self.files.items():
            # Get some attributes.
            data_format = attribs['data_sample_enc']
            endian = attribs['endian']
            count = attribs['sample_count']
            size = attribs['sample_size']
            non_normalized_samples = attribs['non_normalized_samples']
            dtype = self.dtypes[data_format]
            file = os.path.join(self.path, file)
            # Load the data. This data has previously been unpacked by
            # Madagascar.
            data = np.load(file + '.npy').ravel()
            data = np.require(data, dtype)
            # Load the packed data.
            with open(file, 'rb') as f:
                # Jump to the beginning of the data.
                f.seek(3200 + 400 + 240)
                packed_data = f.read(count * size)
            # The pack functions all write to file objects.
            f = io.BytesIO()
            # Pack the data.
            DATA_SAMPLE_FORMAT_PACK_FUNCTIONS[data_format](f, data, endian)
            # Read again.0.
            f.seek(0, 0)
            new_packed_data = f.read()
            # Check the length.
            self.assertEqual(len(packed_data), len(new_packed_data))
            if len(non_normalized_samples) == 0:
                # The packed data should be totally identical.
                self.assertEqual(packed_data, new_packed_data)
            else:
                # Some test files contain non normalized IBM floating point
                # data. These cannot be reproduced exactly.
                # Just a sanity check to be sure it is only IBM floating point
                # data that does not work completely.
                self.assertEqual(data_format, 1)

                # Read the data as uint8 to be able to directly access the
                # different bytes.
                # Original data.
                packed_data = from_buffer(packed_data, np.uint8)
                # Newly written.
                new_packed_data = from_buffer(new_packed_data, np.uint8)

                # Figure out the non normalized fractions in the original data
                # because these cannot be compared directly.
                # Get the position of the first byte of the fraction depending
                # on the endianness.
                if endian == '>':
                    start = 1
                else:
                    start = 2
                # The first byte of the fraction.
                first_fraction_byte_old = packed_data[start::4]
                # First get all zeros in the original data because zeros have
                # to be treated differently.
                zeros = np.where(data == 0)[0]
                # Create a copy and set the zeros to a high number to be able
                # to find all non normalized numbers.
                fraction_copy = first_fraction_byte_old.copy()
                fraction_copy[zeros] = 255
                # Normalized numbers will have no zeros in the first 4 bit of
                # the fraction. This means that the most significant byte of
                # the fraction has to be at least 16 for it to be normalized.
                non_normalized = np.where(fraction_copy < 16)[0]

                # Sanity check if the file data and the calculated data are the
                # same.
                np.testing.assert_array_equal(non_normalized,
                                              np.array(non_normalized_samples))

                # Test all other parts of the packed data. Set dtype to int32
                # to get 4 byte numbers.
                packed_data_copy = packed_data.copy()
                new_packed_data_copy = new_packed_data.copy()
                packed_data_copy.dtype = np.int32
                new_packed_data_copy.dtype = np.int32
                # Equalize the non normalized parts.
                packed_data_copy[non_normalized] = \
                    new_packed_data_copy[non_normalized]
                np.testing.assert_array_equal(packed_data_copy,
                                              new_packed_data_copy)

                # Now check the non normalized parts if they are almost the
                # same.
                data = data[non_normalized]
                # Unpack the data again.
                new_packed_data.dtype = np.int32
                new_packed_data = new_packed_data[non_normalized]
                length = len(new_packed_data)
                f = io.BytesIO()
                f.write(new_packed_data.tostring())
                f.seek(0, 0)
                new_data = DATA_SAMPLE_FORMAT_UNPACK_FUNCTIONS[1](
                    f, length, endian)
                f.close()
                packed_data.dtype = np.int32
                packed_data = packed_data[non_normalized]
                length = len(packed_data)
                f = io.BytesIO()
                f.write(packed_data.tostring())
                f.seek(0, 0)
                old_data = DATA_SAMPLE_FORMAT_UNPACK_FUNCTIONS[1](
                    f, length, endian)
                f.close()
                # This works because the normalized and the non normalized IBM
                # floating point numbers will be close enough for the internal
                # IEEE representation to be identical.
                np.testing.assert_array_equal(data, new_data)
                np.testing.assert_array_equal(data, old_data)

    def test_pack_and_unpack_ibm_float(self):
        """
        Packing and unpacking IBM floating points might yield some inaccuracies
        due to floating point rounding errors.
        This test tests a large number of random floating point numbers.
        """
        # Some random seeds.
        seeds = [1234, 592, 459482, 6901, 0, 7083, 68349]
        endians = ['<', '>']
        # Loop over all combinations.
        for seed in seeds:
            # Generate 50000 random floats from -10000 to +10000.
            np.random.seed(seed)
            data = 200000.0 * np.random.ranf(50000) - 100000.0
            # Convert to float64 in case native floats are different to be
            # able to utilize double precision.
            data = np.require(data, np.float64)
            # Loop over little and big endian.
            for endian in endians:
                # Pack.
                f = io.BytesIO()
                DATA_SAMPLE_FORMAT_PACK_FUNCTIONS[1](f, data, endian)
                # Jump to beginning and read again.
                f.seek(0, 0)
                new_data = DATA_SAMPLE_FORMAT_UNPACK_FUNCTIONS[1](
                    f, len(data), endian)
                f.close()
                # A relative tolerance of 1E-6 is considered good enough.
                rms1 = rms(data, new_data)
                self.assertEqual(True, rms1 < 1E-6)

    def test_pack_and_unpack_very_small_ibm_floats(self):
        """
        The same test as test_packAndUnpackIBMFloat just for small numbers
        because they might suffer more from the inaccuracies.
        """
        # Some random seeds.
        seeds = [123, 1592, 4482, 601, 1, 783, 6849]
        endians = ['<', '>']
        # Loop over all combinations.
        for seed in seeds:
            # Generate 50000 random floats from -10000 to +10000.
            np.random.seed(seed)
            data = 1E-5 * np.random.ranf(50000)
            # Convert to float64 in case native floats are different to be
            # able to utilize double precision.
            data = np.require(data, np.float64)
            # Loop over little and big endian.
            for endian in endians:
                # Pack.
                f = io.BytesIO()
                DATA_SAMPLE_FORMAT_PACK_FUNCTIONS[1](f, data, endian)
                # Jump to beginning and read again.
                f.seek(0, 0)
                new_data = DATA_SAMPLE_FORMAT_UNPACK_FUNCTIONS[1](
                    f, len(data), endian)
                f.close()
                # A relative tolerance of 1E-6 is considered good enough.
                rms1 = rms(data, new_data)
                self.assertEqual(True, rms1 < 1E-6)

    def test_pack_and_unpack_ibm_special_cases(self):
        """
        Tests the packing and unpacking of several powers of 16 which are
        problematic because they need separate handling in the algorithm.
        """
        endians = ['>', '<']
        # Create the first 10 powers of 16.
        data = []
        for i in range(10):
            data.append(16 ** i)
            data.append(-16 ** i)
        data = np.array(data)
        # Convert to float64 in case native floats are different to be
        # able to utilize double precision.
        data = np.require(data, np.float64)
        # Loop over little and big endian.
        for endian in endians:
            # Pack.
            f = io.BytesIO()
            DATA_SAMPLE_FORMAT_PACK_FUNCTIONS[1](f, data, endian)
            # Jump to beginning and read again.
            f.seek(0, 0)
            new_data = DATA_SAMPLE_FORMAT_UNPACK_FUNCTIONS[1](
                f, len(data), endian)
            f.close()
            # Test both.
            np.testing.assert_array_equal(new_data, data)

    def test_read_and_write_binary_file_header(self):
        """
        Reading and writing should not change the binary file header.
        """
        for file, attribs in self.files.items():
            endian = attribs['endian']
            file = os.path.join(self.path, file)
            # Read the file.
            with open(file, 'rb') as f:
                f.seek(3200)
                org_header = f.read(400)
            header = SEGYBinaryFileHeader(header=org_header, endian=endian)
            # The header writes to a file like object.
            new_header = io.BytesIO()
            header.write(new_header)
            new_header.seek(0, 0)
            new_header = new_header.read()
            # Assert the correct length.
            self.assertEqual(len(new_header), 400)
            # Assert the actual header.
            self.assertEqual(org_header, new_header)

    def test_read_and_write_textual_file_header(self):
        """
        Reading and writing should not change the textual file header.
        """
        for file, attribs in self.files.items():
            endian = attribs['endian']
            header_enc = attribs['textual_header_enc']
            file = os.path.join(self.path, file)
            # Read the file.
            with open(file, 'rb') as f:
                org_header = f.read(3200)
                f.seek(0, 0)
                # Initialize an empty SEGY object and set certain attributes.
                segy = SEGYFile()
                segy.endian = endian
                segy.file = f
                segy.textual_header_encoding = None
                # Read the textual header.
                segy._read_textual_header()
                # Assert the encoding and compare with known values.
                self.assertEqual(segy.textual_header_encoding, header_enc)
            # The header writes to a file like object.
            new_header = io.BytesIO()
            with warnings.catch_warnings(record=True):
                segy._write_textual_header(new_header)
            new_header.seek(0, 0)
            new_header = new_header.read()
            # Assert the correct length.
            self.assertEqual(len(new_header), 3200)
            # Patch both headers to not worry about the automatically set
            # values.
            org_header = _patch_header(org_header)
            new_header = _patch_header(new_header)
            # Assert the actual header.
            self.assertEqual(org_header, new_header)

    def test_read_and_write_trace_header(self):
        """
        Reading and writing should not change the trace header.
        """
        for file, attribs in self.files.items():
            endian = attribs['endian']
            file = os.path.join(self.path, file)
            # Read the file.
            with open(file, 'rb') as f:
                f.seek(3600)
                org_header = f.read(240)
            header = SEGYTraceHeader(header=org_header, endian=endian)
            # The header writes to a file like object.
            new_header = io.BytesIO()
            header.write(new_header)
            new_header.seek(0, 0)
            new_header = new_header.read()
            # Assert the correct length.
            self.assertEqual(len(new_header), 240)
            # Assert the actual header.
            self.assertEqual(org_header, new_header)

    def test_read_and_write_segy(self, headonly=False):
        """
        Reading and writing again should not change a file.
        """
        for file, attribs in self.files.items():
            file = os.path.join(self.path, file)
            non_normalized_samples = attribs['non_normalized_samples']
            # Read the file.
            with open(file, 'rb') as f:
                org_data = f.read()
            segy_file = _read_segy(file, headonly=headonly)
            with NamedTemporaryFile() as tf:
                out_file = tf.name
                with warnings.catch_warnings(record=True):
                    segy_file.write(out_file)
                # Read the new file again.
                with open(out_file, 'rb') as f:
                    new_data = f.read()
            # The two files should have the same length.
            self.assertEqual(len(org_data), len(new_data))
            # Replace the not normalized samples. The not normalized
            # samples are already tested in test_packSEGYData and therefore not
            # tested again here.
            if len(non_normalized_samples) != 0:
                # Convert to 4 byte integers. Any 4 byte numbers work.
                org_data = from_buffer(org_data, np.int32)
                new_data = from_buffer(new_data, np.int32)
                # Skip the header (4*960 bytes) and replace the non normalized
                # data samples.
                org_data[960:][non_normalized_samples] = \
                    new_data[960:][non_normalized_samples]
                # Create strings again.
                org_data = org_data.tostring()
                new_data = new_data.tostring()
            # Just patch both headers - this tests something different.
            org_data = _patch_header(org_data)
            new_data = _patch_header(new_data)
            # Always write the SEGY File revision number!
            # org_data[3500:3502] = new_data[3500:3502]
            # Test the identity without the SEGY revision number
            self.assertEqual(org_data[:3500], new_data[:3500])
            self.assertEqual(org_data[3502:], new_data[3502:])

    def test_read_and_write_segy_headonly(self):
        """
        Reading with headonly=True and writing again should not change a file.
        """
        self.test_read_and_write_segy(headonly=True)

    def test_unpack_binary_file_header(self):
        """
        Compares some values of the binary header with values read with
        SeisView 2 by the DMNG.
        """
        file = os.path.join(self.path, '1.sgy_first_trace')
        segy = _read_segy(file)
        header = segy.binary_file_header
        # Compare the values.
        self.assertEqual(header.job_identification_number, 0)
        self.assertEqual(header.line_number, 0)
        self.assertEqual(header.reel_number, 0)
        self.assertEqual(header.number_of_data_traces_per_ensemble, 24)
        self.assertEqual(header.number_of_auxiliary_traces_per_ensemble, 0)
        self.assertEqual(header.sample_interval_in_microseconds, 250)
        self.assertEqual(
            header.sample_interval_in_microseconds_of_original_field_recording,
            250)
        self.assertEqual(header.number_of_samples_per_data_trace, 8000)
        self.assertEqual(
            header.
            number_of_samples_per_data_trace_for_original_field_recording,
            8000)
        self.assertEqual(header.data_sample_format_code, 2)
        self.assertEqual(header.ensemble_fold, 0)
        self.assertEqual(header.trace_sorting_code, 1)
        self.assertEqual(header.vertical_sum_code, 0)
        self.assertEqual(header.sweep_frequency_at_start, 0)
        self.assertEqual(header.sweep_frequency_at_end, 0)
        self.assertEqual(header.sweep_length, 0)
        self.assertEqual(header.sweep_type_code, 0)
        self.assertEqual(header.trace_number_of_sweep_channel, 0)
        self.assertEqual(header.sweep_trace_taper_length_in_ms_at_start, 0)
        self.assertEqual(header.sweep_trace_taper_length_in_ms_at_end, 0)
        self.assertEqual(header.taper_type, 0)
        self.assertEqual(header.correlated_data_traces, 0)
        self.assertEqual(header.binary_gain_recovered, 0)
        self.assertEqual(header.amplitude_recovery_method, 0)
        self.assertEqual(header.measurement_system, 0)
        self.assertEqual(header.impulse_signal_polarity, 0)
        self.assertEqual(header.vibratory_polarity_code, 0)
        self.assertEqual(
            header.number_of_3200_byte_ext_file_header_records_following,
            0)

    def test_unpack_trace_header(self):
        """
        Compares some values of the first trace header with values read with
        SeisView 2 by the DMNG.
        """
        file = os.path.join(self.path, '1.sgy_first_trace')
        segy = _read_segy(file)
        header = segy.traces[0].header
        # Compare the values.
        self.assertEqual(header.trace_sequence_number_within_line, 0)
        self.assertEqual(header.trace_sequence_number_within_segy_file, 0)
        self.assertEqual(header.original_field_record_number, 1)
        self.assertEqual(header.trace_number_within_the_original_field_record,
                         1)
        self.assertEqual(header.energy_source_point_number, 0)
        self.assertEqual(header.ensemble_number, 0)
        self.assertEqual(header.trace_number_within_the_ensemble, 0)
        self.assertEqual(header.trace_identification_code, 1)
        self.assertEqual(
            header.number_of_vertically_summed_traces_yielding_this_trace,
            5)
        self.assertEqual(
            header.number_of_horizontally_stacked_traces_yielding_this_trace,
            0)
        self.assertEqual(header.data_use, 0)
        self.assertEqual(getattr(
            header, 'distance_from_center_of_the_' +
            'source_point_to_the_center_of_the_receiver_group'), 0)
        self.assertEqual(header.receiver_group_elevation, 0)
        self.assertEqual(header.surface_elevation_at_source, 0)
        self.assertEqual(header.source_depth_below_surface, 0)
        self.assertEqual(header.datum_elevation_at_receiver_group, 0)
        self.assertEqual(header.datum_elevation_at_source, 0)
        self.assertEqual(header.water_depth_at_source, 0)
        self.assertEqual(header.water_depth_at_group, 0)
        self.assertEqual(
            header.scalar_to_be_applied_to_all_elevations_and_depths, -100)
        self.assertEqual(header.scalar_to_be_applied_to_all_coordinates, -100)
        self.assertEqual(header.source_coordinate_x, 0)
        self.assertEqual(header.source_coordinate_y, 0)
        self.assertEqual(header.group_coordinate_x, 300)
        self.assertEqual(header.group_coordinate_y, 0)
        self.assertEqual(header.coordinate_units, 0)
        self.assertEqual(header.weathering_velocity, 0)
        self.assertEqual(header.subweathering_velocity, 0)
        self.assertEqual(header.uphole_time_at_source_in_ms, 0)
        self.assertEqual(header.uphole_time_at_group_in_ms, 0)
        self.assertEqual(header.source_static_correction_in_ms, 0)
        self.assertEqual(header.group_static_correction_in_ms, 0)
        self.assertEqual(header.total_static_applied_in_ms, 0)
        self.assertEqual(header.lag_time_A, 0)
        self.assertEqual(header.lag_time_B, 0)
        self.assertEqual(header.delay_recording_time, -100)
        self.assertEqual(header.mute_time_start_time_in_ms, 0)
        self.assertEqual(header.mute_time_end_time_in_ms, 0)
        self.assertEqual(header.number_of_samples_in_this_trace, 8000)
        self.assertEqual(header.sample_interval_in_ms_for_this_trace, 250)
        self.assertEqual(header.gain_type_of_field_instruments, 0)
        self.assertEqual(header.instrument_gain_constant, 24)
        self.assertEqual(header.instrument_early_or_initial_gain, 0)
        self.assertEqual(header.correlated, 0)
        self.assertEqual(header.sweep_frequency_at_start, 0)
        self.assertEqual(header.sweep_frequency_at_end, 0)
        self.assertEqual(header.sweep_length_in_ms, 0)
        self.assertEqual(header.sweep_type, 0)
        self.assertEqual(header.sweep_trace_taper_length_at_start_in_ms, 0)
        self.assertEqual(header.sweep_trace_taper_length_at_end_in_ms, 0)
        self.assertEqual(header.taper_type, 0)
        self.assertEqual(header.alias_filter_frequency, 1666)
        self.assertEqual(header.alias_filter_slope, 0)
        self.assertEqual(header.notch_filter_frequency, 0)
        self.assertEqual(header.notch_filter_slope, 0)
        self.assertEqual(header.low_cut_frequency, 0)
        self.assertEqual(header.high_cut_frequency, 0)
        self.assertEqual(header.low_cut_slope, 0)
        self.assertEqual(header.high_cut_slope, 0)
        self.assertEqual(header.year_data_recorded, 2005)
        self.assertEqual(header.day_of_year, 353)
        self.assertEqual(header.hour_of_day, 15)
        self.assertEqual(header.minute_of_hour, 7)
        self.assertEqual(header.second_of_minute, 54)
        self.assertEqual(header.time_basis_code, 0)
        self.assertEqual(header.trace_weighting_factor, 0)
        self.assertEqual(
            header.geophone_group_number_of_roll_switch_position_one, 2)
        self.assertEqual(header.geophone_group_number_of_trace_number_one, 2)
        self.assertEqual(header.geophone_group_number_of_last_trace, 0)
        self.assertEqual(header.gap_size, 0)
        self.assertEqual(header.over_travel_associated_with_taper, 0)
        self.assertEqual(
            header.x_coordinate_of_ensemble_position_of_this_trace, 0)
        self.assertEqual(
            header.y_coordinate_of_ensemble_position_of_this_trace, 0)
        self.assertEqual(
            header.for_3d_poststack_data_this_field_is_for_in_line_number, 0)
        self.assertEqual(
            header.for_3d_poststack_data_this_field_is_for_cross_line_number,
            0)
        self.assertEqual(header.shotpoint_number, 0)
        self.assertEqual(
            header.scalar_to_be_applied_to_the_shotpoint_number, 0)
        self.assertEqual(header.trace_value_measurement_unit, 0)
        self.assertEqual(header.transduction_constant_mantissa, 0)
        self.assertEqual(header.transduction_constant_exponent, 0)
        self.assertEqual(header.transduction_units, 0)
        self.assertEqual(header.device_trace_identifier, 0)
        self.assertEqual(header.scalar_to_be_applied_to_times, 0)
        self.assertEqual(header.source_type_orientation, 0)
        self.assertEqual(header.source_energy_direction_mantissa, 0)
        self.assertEqual(header.source_energy_direction_exponent, 0)
        self.assertEqual(header.source_measurement_mantissa, 0)
        self.assertEqual(header.source_measurement_exponent, 0)
        self.assertEqual(header.source_measurement_unit, 0)

    def test_read_bytes_io(self):
        """
        Tests reading from BytesIO instances.
        """
        # 1
        file = os.path.join(self.path, 'example.y_first_trace')
        with open(file, 'rb') as f:
            data = f.read()
        st = _read_segy(io.BytesIO(data))
        self.assertEqual(len(st.traces[0].data), 500)
        # 2
        file = os.path.join(self.path, 'ld0042_file_00018.sgy_first_trace')
        with open(file, 'rb') as f:
            data = f.read()
        st = _read_segy(io.BytesIO(data))
        self.assertEqual(len(st.traces[0].data), 2050)
        # 3
        file = os.path.join(self.path, '1.sgy_first_trace')
        with open(file, 'rb') as f:
            data = f.read()
        st = _read_segy(io.BytesIO(data))
        self.assertEqual(len(st.traces[0].data), 8000)
        # 4
        file = os.path.join(self.path, '00001034.sgy_first_trace')
        with open(file, 'rb') as f:
            data = f.read()
        st = _read_segy(io.BytesIO(data))
        self.assertEqual(len(st.traces[0].data), 2001)
        # 5
        file = os.path.join(self.path, 'planes.segy_first_trace')
        with open(file, 'rb') as f:
            data = f.read()
        st = _read_segy(io.BytesIO(data))
        self.assertEqual(len(st.traces[0].data), 512)

    def test_iterative_reading(self):
        """
        Tests iterative reading.
        """
        # Read normally.
        filename = os.path.join(self.path, 'example.y_first_trace')
        st = obspy.read(filename, unpack_trace_headers=True)

        # Read iterative.
        ist = [_i for _i in iread_segy(filename, unpack_headers=True)]

        del ist[0].stats.segy.textual_file_header
        del ist[0].stats.segy.binary_file_header
        del ist[0].stats.segy.textual_file_header_encoding
        del ist[0].stats.segy.data_encoding
        del ist[0].stats.segy.endian

        self.assertEqual(st.traces, ist)

    def test_revision_number_in_binary_file_header(self):
        """
        It is a bit awkward but it is encoded in a 16 bit number with the
        dot assumed to be between the first and second byte. So the hex 16
        representation is 0x0100.
        """
        tr = obspy.read()[0]
        tr.data = np.float32(tr.data)
        for endian in ("<", ">"):
            _tr = tr.copy()
            with io.BytesIO() as buf:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    _tr.write(buf, format="segy", byteorder=endian)
                buf.seek(0, 0)
                data = buf.read()
            # Result differs depending on byte order.
            if endian == "<":
                self.assertEqual(data[3200:3600][-100:-98], b"\x00\x01")
            else:
                self.assertEqual(data[3200:3600][-100:-98], b"\x01\x00")

    def test_textual_header_has_the_right_fields_at_the_end(self):
        """
        Needs to have the version number and a magical string at the end.
        """
        tr = obspy.read()[0]
        tr.data = np.float32(tr.data)

        # If both are missing they will be replaced.
        with io.BytesIO() as buf:
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                tr.write(buf, format="segy")
            self.assertEqual(
                len([_i for _i in w
                     if _i.category is SEGYInvalidTextualHeaderWarning]), 0)
            buf.seek(0, 0)
            data = buf.read()

        # Make sure the textual header has the required fields.
        revision_number = data[:3200][-160:-146].decode()
        end_header_mark = data[:3200][-80:-58].decode()
        self.assertEqual(revision_number, "C39 SEG Y REV1")
        self.assertEqual(end_header_mark, "C40 END TEXTUAL HEADER")

        # An alternate end marker is accepted - no warning will be raised in
        # this case.
        # If both are missing they will be replaced.
        st = obspy.Stream(traces=[tr.copy()])
        st.stats = AttribDict()
        # Write the alternate file header.
        st.stats.textual_file_header = \
            b" " * (3200 - 80) + b"C40 END EBCDIC        " + b" " * 58
        with io.BytesIO() as buf:
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                st.write(buf, format="segy", textual_header_encoding="EBCDIC")
            self.assertEqual(
                len([_i for _i in w
                     if _i.category is SEGYInvalidTextualHeaderWarning]), 0)
            buf.seek(0, 0)
            data = buf.read()

        # Make sure the textual header has the required fields.
        revision_number = data[:3200][-160:-146].decode("EBCDIC-CP-BE")
        end_header_mark = data[:3200][-80:-58].decode("EBCDIC-CP-BE")
        self.assertEqual(revision_number, "C39 SEG Y REV1")
        self.assertEqual(end_header_mark, "C40 END EBCDIC        ")

        # Putting the correct values will raise no warning and leave the
        # values.
        st = obspy.Stream(traces=[tr.copy()])
        st.stats = AttribDict()
        # Write the alternate file header.
        st.stats.textual_file_header =  \
            b" " * (3200 - 160) + b"C39 SEG Y REV1" + b" " * 66 + \
            b"C40 END TEXTUAL HEADER" + b" " * 58
        with io.BytesIO() as buf:
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                st.write(buf, format="segy")
            self.assertEqual(
                len([_i for _i in w
                     if _i.category is SEGYInvalidTextualHeaderWarning]), 0)
            buf.seek(0, 0)
            data = buf.read()

        # Make sure the textual header has the required fields.
        revision_number = data[:3200][-160:-146].decode()
        end_header_mark = data[:3200][-80:-58].decode()
        self.assertEqual(revision_number, "C39 SEG Y REV1")
        self.assertEqual(end_header_mark, "C40 END TEXTUAL HEADER")

        # Putting a wrong revision number will raise a warning, but it will
        # still be written.
        st = obspy.Stream(traces=[tr.copy()])
        st.stats = AttribDict()
        # Write the alternate file header.
        st.stats.textual_file_header = \
            b" " * (3200 - 160) + b"ABCDEFGHIJKLMN" + b" " * 146
        with io.BytesIO() as buf:
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                st.write(buf, format="segy")
            w = [_i for _i in w
                 if _i.category is SEGYInvalidTextualHeaderWarning]
            self.assertEqual(len(w), 1)
            self.assertEqual(
                w[0].message.args[0],
                "The revision number in the textual header should be set as "
                "'C39 SEG Y REV1' for a fully valid SEG-Y file. It is set to "
                "'ABCDEFGHIJKLMN' which will be written to the file. Please "
                "change it if you want a fully valid file.")
            buf.seek(0, 0)
            data = buf.read()

        revision_number = data[:3200][-160:-146].decode()
        end_header_mark = data[:3200][-80:-58].decode()
        # The field is still written.
        self.assertEqual(revision_number, "ABCDEFGHIJKLMN")
        self.assertEqual(end_header_mark, "C40 END TEXTUAL HEADER")

        # Same with the end header mark.
        st = obspy.Stream(traces=[tr.copy()])
        st.stats = AttribDict()
        # Write the alternate file header.
        st.stats.textual_file_header = \
            b" " * (3200 - 80) + b"ABCDEFGHIJKLMNOPQRSTUV" + b" " * 58
        with io.BytesIO() as buf:
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                st.write(buf, format="segy")
            w = [_i for _i in w
                 if _i.category is SEGYInvalidTextualHeaderWarning]
            self.assertEqual(len(w), 1)
            self.assertEqual(
                w[0].message.args[0],
                "The end header mark in the textual header should be set as "
                "'C40 END TEXTUAL HEADER' or as 'C40 END EBCDIC        ' for "
                "a fully valid SEG-Y file. It is "
                "set to 'ABCDEFGHIJKLMNOPQRSTUV' which will be written to the "
                "file. Please change it if you want a fully valid file.")
            buf.seek(0, 0)
            data = buf.read()

        revision_number = data[:3200][-160:-146].decode()
        end_header_mark = data[:3200][-80:-58].decode()
        # The field is still written.
        self.assertEqual(revision_number, "C39 SEG Y REV1")
        self.assertEqual(end_header_mark, "ABCDEFGHIJKLMNOPQRSTUV")

    def test_packing_raises_nice_error_messages(self):
        """
        SEG-Y is fairly restrictive in what ranges are allowed for its header
        values. Thus we attempt to raise nice and helpful error messages.
        """
        tr = obspy.read()[0]
        tr.data = np.float32(np.zeros(3))
        # mock the number of traces per ensemble with an invalid number to
        # trigger error message. we have a more meaningful message for number
        # of samples now, so we can't use that anymore for testing any
        # arbitrary invalid binary header value
        with mock.patch.object(
                SEGYBinaryFileHeader, 'number_of_data_traces_per_ensemble',
                create=True, new_callable=mock.PropertyMock,
                return_value=100000):
            with io.BytesIO() as buf:
                with self.assertRaises(ValueError) as err:
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore")
                        tr.write(buf, format="segy")
        self.assertEqual(
            err.exception.args[0],
            "Failed to pack header value `number_of_data_traces_per_ensemble` "
            "(100000) with format `>h` due to: `'h' format requires -32768 <="
            " number <= 32767`")


def rms(x, y):
    """
    Normalized RMS

    Taken from the mtspec library:
    https://github.com/krischer/mtspec
    """
    return np.sqrt(((x - y) ** 2).mean() / (x ** 2).mean())


def suite():
    return unittest.makeSuite(SEGYTestCase, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# -*- coding: utf-8 -*-
"""
Module for ObsPy's default colormaps.

Overview of provided colormaps:
===============================

The following colormaps can be imported like..

    >>> from obspy.imaging.cm import viridis_r

List of all colormaps:

    * `viridis`_
    * `viridis_r`_
    * `viridis_white`_
    * `viridis_white_r`_
    * obspy_sequential (alias for `viridis`_)
    * obspy_sequential_r (alias for `viridis_r`_)
    * obspy_divergent (alias for matplotlib's RdBu_r)
    * obspy_divergent_r (alias for matplotlib's RdBu)
    * `pqlx`_

.. plot::

    from obspy.imaging.cm import _colormap_plot_overview
    _colormap_plot_overview()

viridis
-------

"viridis" is matplotlib's new default colormap from version 2.0 onwards and is
based on a design by Eric Firing (@efiring, see
http://thread.gmane.org/gmane.comp.python.matplotlib.devel/13522/focus=13542).

    >>> from obspy.imaging.cm import viridis

.. plot::

    from obspy.imaging.cm import viridis as cmap
    from obspy.imaging.cm import _colormap_plot_cwt as plot
    plot([cmap])

.. plot::

    from obspy.imaging.cm import viridis as cmap
    from obspy.imaging.cm import _colormap_plot_array_response as plot
    plot([cmap])

.. plot::

    from obspy.imaging.cm import viridis as cmap
    from obspy.imaging.cm import _colormap_plot_ppsd as plot
    plot([cmap])

.. plot::

    from obspy.imaging.cm import viridis as cmap
    from obspy.imaging.cm import _colormap_plot_similarity as plot
    plot([cmap])

.. plot::

    from obspy.imaging.cm import viridis as cmap
    from obspy.imaging.cm import _colormap_plot_beamforming_time as plot
    plot([cmap])

viridis_r
---------

Reversed version of viridis.

    >>> from obspy.imaging.cm import viridis_r

.. plot::

    from obspy.imaging.cm import viridis_r as cmap
    from obspy.imaging.cm import _colormap_plot_cwt as plot
    plot([cmap])

.. plot::

    from obspy.imaging.cm import viridis_r as cmap
    from obspy.imaging.cm import _colormap_plot_array_response as plot
    plot([cmap])

.. plot::

    from obspy.imaging.cm import viridis_r as cmap
    from obspy.imaging.cm import _colormap_plot_ppsd as plot
    plot([cmap])

.. plot::

    from obspy.imaging.cm import viridis_r as cmap
    from obspy.imaging.cm import _colormap_plot_similarity as plot
    plot([cmap])

.. plot::

    from obspy.imaging.cm import viridis_r as cmap
    from obspy.imaging.cm import _colormap_plot_beamforming_time as plot
    plot([cmap])

viridis_white
-------------

"viridis_white" is a modified version of "viridis" that goes to white instead
of yellow in the end. Although it remains perceptually uniform, the light
colors are a bit more difficult to distinguish than yellow in the original
viridis. It is useful for printing because one end of the colorbar can merge
with a white background (by M Meschede).

    >>> from obspy.imaging.cm import viridis_white

.. plot::

    from obspy.imaging.cm import viridis_white as cmap
    from obspy.imaging.cm import _colormap_plot_cwt as plot
    plot([cmap])

.. plot::

    from obspy.imaging.cm import viridis_white as cmap
    from obspy.imaging.cm import _colormap_plot_array_response as plot
    plot([cmap])

.. plot::

    from obspy.imaging.cm import viridis_white as cmap
    from obspy.imaging.cm import _colormap_plot_ppsd as plot
    plot([cmap])

.. plot::

    from obspy.imaging.cm import viridis_white as cmap
    from obspy.imaging.cm import _colormap_plot_similarity as plot
    plot([cmap])

.. plot::

    from obspy.imaging.cm import viridis_white as cmap
    from obspy.imaging.cm import _colormap_plot_beamforming_time as plot
    plot([cmap])

viridis_white_r
---------------

Reversed version of viridis_white.

    >>> from obspy.imaging.cm import viridis_white_r

.. plot::

    from obspy.imaging.cm import viridis_white_r as cmap
    from obspy.imaging.cm import _colormap_plot_cwt as plot
    plot([cmap])

.. plot::

    from obspy.imaging.cm import viridis_white_r as cmap
    from obspy.imaging.cm import _colormap_plot_array_response as plot
    plot([cmap])

.. plot::

    from obspy.imaging.cm import viridis_white_r as cmap
    from obspy.imaging.cm import _colormap_plot_ppsd as plot
    plot([cmap])

.. plot::

    from obspy.imaging.cm import viridis_white_r as cmap
    from obspy.imaging.cm import _colormap_plot_similarity as plot
    plot([cmap])

.. plot::

    from obspy.imaging.cm import viridis_white_r as cmap
    from obspy.imaging.cm import _colormap_plot_beamforming_time as plot
    plot([cmap])

pqlx
----

Colormap defined and used in PQLX (see [McNamara2004]_).

    >>> from obspy.imaging.cm import pqlx

.. plot::

    from obspy.imaging.cm import pqlx as cmap
    from obspy.imaging.cm import _colormap_plot_ppsd as plot
    plot([cmap])

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA @UnusedWildImport

import glob
import inspect
import os

import numpy as np
from matplotlib.cm import get_cmap
from matplotlib.colors import LinearSegmentedColormap
from matplotlib.colors import ListedColormap


def _get_cmap(file_name, lut=None, reverse=False):
    """
    Load a :class:`~matplotlib.colors.LinearSegmentedColormap` from
    `segmentdata` dictionary saved as numpy compressed binary data.

    :type file_name: str
    :param file_name: Name of colormap to load, same as filename in
                      `obspy/imaging/data`. The type of colormap data is
                      determined from the extension: .npz assumes the file
                      contains colorbar segments (segmented colormap).
                      '*.npy' assumes the file contains a simple array of RGB
                      values with size [ncolors, 3].
    :type lut: int
    :param lut: Specifies the number of discrete color values in the
                segmented colormap. Only used for segmented colormap
                `None` to use matplotlib default value (continuous colormap).
    :type reverse: bool
    :param reverse: Whether to return the specified colormap reverted.
    :rtype: :class:`~matplotlib.colors.LinearSegmentedColormap`
    """
    file_name = file_name.strip()
    name, suffix = os.path.splitext(file_name)
    directory = os.path.dirname(os.path.abspath(
        inspect.getfile(inspect.currentframe())))
    directory = os.path.join(directory, "data")
    full_path = os.path.join(directory, file_name)
    # check if it is npz -> segmented colormap or npy -> listed colormap
    # do it like matplotlib, append "_r" to reverted versions
    if reverse:
        name += "_r"

    if suffix == '.npz':
        # segmented colormap
        data = dict(np.load(full_path))
        if reverse:
            data_r = {}
            for key, val in data.items():
                # copied from matplotlib source,
                # cm.py@f7a578656abc2b2c13 line 47
                data_r[key] = [(1.0 - x, y1, y0) for x, y0, y1 in
                               reversed(val)]
            data = data_r
        kwargs = lut and {"N": lut} or {}
        cmap = LinearSegmentedColormap(name=name, segmentdata=data, **kwargs)
    elif suffix == '.npy':
        # listed colormap
        data = np.load(full_path)
        if reverse:
            data = data[::-1]
        cmap = ListedColormap(data, name=name)
    else:
        raise ValueError('file suffix {} not recognized.'.format(suffix))

    return cmap


def _get_all_cmaps():
    """
    Return all colormaps in "obspy/imaging/data" directory, including reversed
    versions.

    :rtype: dict
    """
    cmaps = {}
    cm_file_pattern = os.path.join(
        os.path.abspath(os.path.dirname(
            inspect.getfile(inspect.currentframe()))),
        "data", "*.np[yz]")
    for filename in glob.glob(cm_file_pattern):
        filename = os.path.basename(filename)
        for reverse in (True, False):
            # don't add a reversed version for PQLX colormap
            if filename == "pqlx.npz" and reverse:
                continue
            cmap = _get_cmap(filename, reverse=reverse)
            cmaps[cmap.name] = cmap
    return cmaps


# inject all colormaps into namespace
_globals = globals()
_globals.update(_get_all_cmaps())

obspy_sequential = _globals["viridis"]
obspy_sequential_r = _globals["viridis_r"]
obspy_divergent = get_cmap("RdBu_r")
obspy_divergent_r = get_cmap("RdBu")
pqlx = _get_cmap("pqlx.npz")


def _colormap_plot_overview(colormap_names=(
        "viridis", "obspy_sequential", "viridis_white", "viridis_r",
        "obspy_sequential_r", "viridis_white_r", "obspy_divergent",
        "obspy_divergent_r", "pqlx")):
    """
    Overview bar plot, adapted after
    http://scipy-cookbook.readthedocs.org/items/Matplotlib_Show_colormaps.html.
    """
    import matplotlib.pyplot as plt
    import importlib
    cm = importlib.import_module("obspy.imaging.cm")
    plt.rc('text', usetex=False)
    a = np.outer(np.ones(1000), np.linspace(0, 1, 1000))
    fig = plt.figure(figsize=(12, 6))
    fig.subplots_adjust(top=0.9, bottom=0.15, left=0.11, right=0.89)
    extent = (0, 1, 0, 1)
    for i, name in enumerate(colormap_names):
        cmap = getattr(cm, name)
        ax = fig.add_subplot(len(colormap_names), 1, i + 1)
        ax.imshow(a, aspect='auto', cmap=cmap, origin="lower", extent=extent,
                  interpolation="nearest")
        ax.set_ylabel(name, family="monospace", fontsize="large", ha="right",
                      rotation="horizontal")
    for ax in fig.axes:
        plt.setp(ax.get_yticklabels(), visible=False)
        ax.yaxis.set_ticks_position("none")
    for ax in fig.axes[:-1]:
        plt.setp(ax.get_xticklabels(), visible=False)
    fig.tight_layout()
    plt.show()


def _colormap_plot_ppsd(cmaps):
    """
    Plot for illustrating colormaps: PPSD.

    :param cmaps: list of :class:`~matplotlib.colors.Colormap`
    :rtype: None
    """
    import matplotlib.pyplot as plt
    from obspy import read
    from obspy.signal import PPSD
    from obspy.io.xseed import Parser
    st = read("https://examples.obspy.org/BW.KW1..EHZ.D.2011.037")
    st += read("https://examples.obspy.org/BW.KW1..EHZ.D.2011.038")
    parser = Parser("https://examples.obspy.org/dataless.seed.BW_KW1")
    ppsd = PPSD(st[0].stats, metadata=parser)
    ppsd.add(st)

    for cmap in cmaps:
        ppsd.plot(cmap=cmap, show=False)
    plt.show()


def _colormap_plot_array_response(cmaps):
    """
    Plot for illustrating colormaps: array response.

    :param cmaps: list of :class:`~matplotlib.colors.Colormap`
    :rtype: None
    """
    import matplotlib.pyplot as plt
    from obspy.signal.array_analysis import array_transff_wavenumber
    # generate array coordinates
    coords = np.array([[10., 60., 0.], [200., 50., 0.], [-120., 170., 0.],
                       [-100., -150., 0.], [30., -220., 0.]])
    # coordinates in km
    coords /= 1000.
    # set limits for wavenumber differences to analyze
    klim = 40.
    kxmin = -klim
    kxmax = klim
    kymin = -klim
    kymax = klim
    kstep = klim / 100.
    # compute transfer function as a function of wavenumber difference
    transff = array_transff_wavenumber(coords, klim, kstep, coordsys='xy')
    # plot
    for cmap in cmaps:
        plt.figure()
        plt.pcolor(np.arange(kxmin, kxmax + kstep * 1.1, kstep) - kstep / 2.,
                   np.arange(kymin, kymax + kstep * 1.1, kstep) - kstep / 2.,
                   transff.T, cmap=cmap)
        plt.colorbar()
        plt.clim(vmin=0., vmax=1.)
        plt.xlim(kxmin, kxmax)
        plt.ylim(kymin, kymax)
    plt.show()


def _colormap_plot_cwt(cmaps):
    """
    Plot for illustrating colormaps: cwt.

    :param cmaps: list of :class:`~matplotlib.colors.Colormap`
    :rtype: None
    """
    import matplotlib.pyplot as plt
    from obspy import read
    from obspy.signal.tf_misfit import cwt
    tr = read()[0]
    npts = tr.stats.npts
    dt = tr.stats.delta
    t = np.linspace(0, dt * npts, npts)
    f_min = 1
    f_max = 50
    scalogram = cwt(tr.data, dt, 8, f_min, f_max)
    x, y = np.meshgrid(
        t, np.logspace(np.log10(f_min), np.log10(f_max), scalogram.shape[0]))
    for cmap in cmaps:
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.pcolormesh(x, y, np.abs(scalogram), cmap=cmap)
        ax.set_xlabel("Time after %s [s]" % tr.stats.starttime)
        ax.set_ylabel("Frequency [Hz]")
        ax.set_yscale('log')
        ax.set_ylim(f_min, f_max)
    plt.show()


def _colormap_plot_similarity(cmaps):
    """
    Plot for illustrating colormaps: similarity matrix.

    :param cmaps: list of :class:`~matplotlib.colors.Colormap`
    :rtype: None
    """
    import matplotlib.pyplot as plt
    from future import standard_library
    standard_library.install_aliases()
    import io
    from urllib.request import urlopen

    url = "https://examples.obspy.org/dissimilarities.npz"
    with io.BytesIO(urlopen(url).read()) as fh, np.load(fh) as data:
        dissimilarity = data['dissimilarity']

    for cmap in cmaps:
        plt.figure(figsize=(6, 5))
        plt.subplot(1, 1, 1)
        plt.imshow(1 - dissimilarity, interpolation='nearest', cmap=cmap)
        plt.xlabel("Event number")
        plt.ylabel("Event number")
        cb = plt.colorbar()
        cb.set_label("Similarity")
    plt.show()


def _get_beamforming_example_stream():
    # Load data
    from obspy import read
    from obspy.core.util import AttribDict
    from obspy.signal.invsim import corn_freq_2_paz
    st = read("https://examples.obspy.org/agfa.mseed")
    # Set PAZ and coordinates for all 5 channels
    st[0].stats.paz = AttribDict({
        'poles': [(-0.03736 - 0.03617j), (-0.03736 + 0.03617j)],
        'zeros': [0j, 0j],
        'sensitivity': 205479446.68601453,
        'gain': 1.0})
    st[0].stats.coordinates = AttribDict({
        'latitude': 48.108589,
        'elevation': 0.450000,
        'longitude': 11.582967})
    st[1].stats.paz = AttribDict({
        'poles': [(-0.03736 - 0.03617j), (-0.03736 + 0.03617j)],
        'zeros': [0j, 0j],
        'sensitivity': 205479446.68601453,
        'gain': 1.0})
    st[1].stats.coordinates = AttribDict({
        'latitude': 48.108192,
        'elevation': 0.450000,
        'longitude': 11.583120})
    st[2].stats.paz = AttribDict({
        'poles': [(-0.03736 - 0.03617j), (-0.03736 + 0.03617j)],
        'zeros': [0j, 0j],
        'sensitivity': 250000000.0,
        'gain': 1.0})
    st[2].stats.coordinates = AttribDict({
        'latitude': 48.108692,
        'elevation': 0.450000,
        'longitude': 11.583414})
    st[3].stats.paz = AttribDict({
        'poles': [(-4.39823 + 4.48709j), (-4.39823 - 4.48709j)],
        'zeros': [0j, 0j],
        'sensitivity': 222222228.10910088,
        'gain': 1.0})
    st[3].stats.coordinates = AttribDict({
        'latitude': 48.108456,
        'elevation': 0.450000,
        'longitude': 11.583049})
    st[4].stats.paz = AttribDict({
        'poles': [(-4.39823 + 4.48709j), (-4.39823 - 4.48709j), (-2.105 + 0j)],
        'zeros': [0j, 0j, 0j],
        'sensitivity': 222222228.10910088,
        'gain': 1.0})
    st[4].stats.coordinates = AttribDict({
        'latitude': 48.108730,
        'elevation': 0.450000,
        'longitude': 11.583157})
    # Instrument correction to 1Hz corner frequency
    paz1hz = corn_freq_2_paz(1.0, damp=0.707)
    st.simulate(paz_remove='self', paz_simulate=paz1hz)
    return st


def _colormap_plot_beamforming_time(cmaps):
    """
    Plot for illustrating colormaps: beamforming.

    :param cmaps: list of :class:`~matplotlib.colors.Colormap`
    :rtype: None
    """
    import matplotlib.pyplot as plt
    import matplotlib.dates as mdates

    from obspy import UTCDateTime
    from obspy.signal.array_analysis import array_processing

    # Execute array_processing
    stime = UTCDateTime("20080217110515")
    etime = UTCDateTime("20080217110545")
    kwargs = dict(
        # slowness grid: X min, X max, Y min, Y max, Slow Step
        sll_x=-3.0, slm_x=3.0, sll_y=-3.0, slm_y=3.0, sl_s=0.03,
        # sliding window properties
        win_len=1.0, win_frac=0.05,
        # frequency properties
        frqlow=1.0, frqhigh=8.0, prewhiten=0,
        # restrict output
        semb_thres=-1e9, vel_thres=-1e9, timestamp='mlabday',
        stime=stime, etime=etime
    )
    st = _get_beamforming_example_stream()
    out = array_processing(st, **kwargs)
    # Plot
    labels = ['rel.power', 'abs.power', 'baz', 'slow']
    xlocator = mdates.AutoDateLocator()
    for cmap in cmaps:
        fig = plt.figure()
        for i, lab in enumerate(labels):
            ax = fig.add_subplot(4, 1, i + 1)
            ax.scatter(out[:, 0], out[:, i + 1], c=out[:, 1], alpha=0.6,
                       edgecolors='none', cmap=cmap)
            ax.set_ylabel(lab)
            ax.set_xlim(out[0, 0], out[-1, 0])
            ax.set_ylim(out[:, i + 1].min(), out[:, i + 1].max())
            ax.xaxis.set_major_locator(xlocator)
            ax.xaxis.set_major_formatter(mdates.AutoDateFormatter(xlocator))
        fig.suptitle('AGFA skyscraper blasting in Munich %s' % (
            stime.strftime('%Y-%m-%d'), ))
        fig.autofmt_xdate()
        fig.subplots_adjust(left=0.15, top=0.95, right=0.95, bottom=0.2,
                            hspace=0)
    plt.show()


def _colormap_plot_beamforming_polar(cmaps):
    """
    Plot for illustrating colormaps: beamforming.

    :param cmaps: list of :class:`~matplotlib.colors.Colormap`
    :rtype: None
    """
    import matplotlib.pyplot as plt
    from matplotlib.colorbar import ColorbarBase
    from matplotlib.colors import Normalize

    from obspy import UTCDateTime
    from obspy.signal.array_analysis import array_processing
    # Execute array_processing
    kwargs = dict(
        # slowness grid: X min, X max, Y min, Y max, Slow Step
        sll_x=-3.0, slm_x=3.0, sll_y=-3.0, slm_y=3.0, sl_s=0.03,
        # sliding window properties
        win_len=1.0, win_frac=0.05,
        # frequency properties
        frqlow=1.0, frqhigh=8.0, prewhiten=0,
        # restrict output
        semb_thres=-1e9, vel_thres=-1e9,
        stime=UTCDateTime("20080217110515"),
        etime=UTCDateTime("20080217110545")
    )
    st = _get_beamforming_example_stream()
    out = array_processing(st, **kwargs)
    # make output human readable, adjust backazimuth to values between 0 and
    # 360
    t, rel_power, abs_power, baz, slow = out.T
    baz[baz < 0.0] += 360
    # choose number of fractions in plot (desirably 360 degree/N is an
    # integer!)
    num = 36
    num2 = 30
    abins = np.arange(num + 1) * 360. / num
    sbins = np.linspace(0, 3, num2 + 1)
    # sum rel power in bins given by abins and sbins
    hist, baz_edges, sl_edges = \
        np.histogram2d(baz, slow, bins=[abins, sbins], weights=rel_power)
    # transform to radian
    baz_edges = np.radians(baz_edges)
    dh = abs(sl_edges[1] - sl_edges[0])
    dw = abs(baz_edges[1] - baz_edges[0])
    for cmap in cmaps:
        # add polar and colorbar axes
        fig = plt.figure(figsize=(8, 8))
        cax = fig.add_axes([0.85, 0.2, 0.05, 0.5])
        ax = fig.add_axes([0.10, 0.1, 0.70, 0.7], polar=True)
        ax.set_theta_direction(-1)
        ax.set_theta_zero_location("N")
        # circle through backazimuth
        for i, row in enumerate(hist):
            ax.bar(left=(i * dw) * np.ones(num2),
                   height=dh * np.ones(num2),
                   width=dw, bottom=dh * np.arange(num2),
                   color=cmap(row / hist.max()))
        ax.set_xticks(np.linspace(0, 2 * np.pi, 4, endpoint=False))
        ax.set_xticklabels(['N', 'E', 'S', 'W'])
        # set slowness limits
        ax.set_ylim(0, 3)
        [i.set_color('grey') for i in ax.get_yticklabels()]
        ColorbarBase(cax, cmap=cmap,
                     norm=Normalize(vmin=hist.min(), vmax=hist.max()))
    plt.show()


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Holds various helper classes to keep the file number manageable.
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA
from future.utils import native_str

from collections import namedtuple

import numpy as np


class SlownessModelError(Exception):
    pass


class TauModelError(Exception):
    pass


SlownessLayer = np.dtype([
    (native_str('top_p'), np.float_),
    (native_str('top_depth'), np.float_),
    (native_str('bot_p'), np.float_),
    (native_str('bot_depth'), np.float_),
])


"""
Holds the ray parameter, time and distance increments, and optionally a
depth, for a ray passing through some layer.
"""
TimeDist = np.dtype([
    (native_str('p'), np.float_),
    (native_str('time'), np.float_),
    (native_str('dist'), np.float_),
    (native_str('depth'), np.float_),
])


"""
Holds the ray parameter, time and distance increments, and optionally a
depth, latitude and longitude for a ray passing through some layer.
"""
TimeDistGeo = np.dtype([
    (native_str('p'), np.float_),
    (native_str('time'), np.float_),
    (native_str('dist'), np.float_),
    (native_str('depth'), np.float_),
    (native_str('lat'), np.float_),
    (native_str('lon'), np.float_)
])


"""
Tracks critical points (discontinuities or reversals in slowness gradient)
within slowness and velocity models.
"""
CriticalDepth = np.dtype([
    (native_str('depth'), np.float_),
    (native_str('vel_layer_num'), np.int_),
    (native_str('p_layer_num'), np.int_),
    (native_str('s_layer_num'), np.int_),
])


class DepthRange:
    """
    Convenience class for storing a depth range. It has a top and a bottom and
    can have an associated ray parameter.
    """
    def __init__(self, top_depth=None, bot_depth=None, ray_param=-1):
        self.top_depth = top_depth
        self.bot_depth = bot_depth
        self.ray_param = ray_param

    def _to_array(self):
        """
        Store all attributes for serialization in a structured array.
        """
        arr = np.empty(3, dtype=np.float_)
        arr[0] = self.top_depth
        arr[1] = self.bot_depth
        arr[2] = self.ray_param
        return arr

    @staticmethod
    def _from_array(arr):
        """
        Create instance object from a structured array used in serialization.
        """
        depth_range = DepthRange()
        depth_range.top_depth = arr[0]
        depth_range.bot_depth = arr[1]
        depth_range.ray_param = arr[2]
        return depth_range


SplitLayerInfo = namedtuple(
    'SplitLayerInfo',
    ['s_mod', 'needed_split', 'moved_sample', 'ray_param']
)


class Arrival(object):
    """
    Convenience class for storing parameters associated with a phase arrival.

    :ivar phase: Phase that generated this arrival
    :vartype phase: :class:`~obspy.taup.seismic_phase.SeismicPhase`
    :ivar distance: Actual distance in degrees
    :vartype distance: float
    :ivar time: Travel time in seconds
    :vartype time: float
    :ivar purist_dist: Purist angular distance (great circle) in radians
    :vartype purist_dist: float
    :ivar ray_param: Ray parameter in seconds per radians
    :vartype ray_param: float
    :ivar name: Phase name
    :vartype name: str
    :ivar purist_name: Phase name changed for true depths
    :vartype purist_name: str
    :ivar source_depth: Source depth in kilometers
    :vartype source_depth: float
    :ivar incident_angle: Angle (in degrees) at which the ray arrives at the
        receiver
    :vartype incident_angle: float
    :ivar takeoff_angle: Angle (in degrees) at which the ray leaves the source
    :vartype takeoff_angle: float
    :ivar pierce: Points pierced by ray
    :vartype pierce: :class:`~numpy.ndarray` (dtype = :const:`~TimeDist`)
    :ivar path: Path taken by ray
    :vartype path: :class:`~numpy.ndarray` (dtype = :const:`~TimeDist`)
    """
    def __init__(self, phase, distance, time, purist_dist, ray_param,
                 ray_param_index, name, purist_name, source_depth,
                 receiver_depth, takeoff_angle=None, incident_angle=None):
        if np.isnan(time):
            raise ValueError('Time cannot be NaN')
        if ray_param_index < 0:
            raise ValueError(
                'ray_param_index cannot be negative: %d' % (ray_param_index, ))

        self.phase = phase
        self.distance = distance
        self.time = time
        self.purist_dist = purist_dist
        self.ray_param = ray_param
        self.ray_param_index = ray_param_index
        self.name = name
        self.purist_name = purist_name
        self.source_depth = source_depth
        self.receiver_depth = receiver_depth
        if takeoff_angle is None:
            self.takeoff_angle = phase.calc_takeoff_angle(ray_param)
        else:
            self.takeoff_angle = takeoff_angle
        if incident_angle is None:
            self.incident_angle = phase.calc_incident_angle(ray_param)
        else:
            self.incident_angle = incident_angle
        self.pierce = None
        self.path = None

    def __str__(self):
        return "%s phase arrival at %.3f seconds" % (self.phase.name,
                                                     self.time)

    @property
    def ray_param_sec_degree(self):
        """
        Return the ray parameter in seconds per degree.
        """
        return self.ray_param * np.pi / 180.0

    @property
    def purist_distance(self):
        """
        Return the purist distance in degrees.
        """
        return self.purist_dist * 180.0 / np.pi



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Functionality for dealing with a single velocity layer.
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA
from future.utils import native_str

import numpy as np


#: The VelocityLayer dtype stores a single layer. An entire velocity model is
#: implemented as an array of layers. The elements are:
#:
#: * ``top_depth``: The top depth of the layer.
#: * ``bot_depth``: The bottom depth of the layer.
#: * ``top_p_velocity``: The compressional (P) wave velocity at the top.
#: * ``bot_p_velocity``: The compressional (P) wave velocity at the bottom.
#: * ``top_s_velocity``: The shear (S) wave velocity at the top.
#: * ``bot_s_velocity``: The shear (S) wave velocity at the bottom.
#: * ``top_density``: The density at the top.
#: * ``bot_density``: The density at the bottom.
#: * ``top_qp``: The P wave attenuation at the top.
#: * ``bot_qp``: The P wave attenuation at the bottom.
#: * ``top_qs``: The S wave attenuation at the top.
#: * ``bot_qs``: The S wave attenuation at the bottom.
VelocityLayer = np.dtype([
    (native_str('top_depth'), np.float_),
    (native_str('bot_depth'), np.float_),
    (native_str('top_p_velocity'), np.float_),
    (native_str('bot_p_velocity'), np.float_),
    (native_str('top_s_velocity'), np.float_),
    (native_str('bot_s_velocity'), np.float_),
    (native_str('top_density'), np.float_),
    (native_str('bot_density'), np.float_),
    (native_str('top_qp'), np.float_),
    (native_str('bot_qp'), np.float_),
    (native_str('top_qs'), np.float_),
    (native_str('bot_qs'), np.float_),
])


def evaluate_velocity_at_bottom(layer, prop):
    """
    Evaluate material properties at bottom of a velocity layer.

    .. seealso:: :func:`evaluate_velocity_at_top`, :func:`evaluate_velocity_at`

    :param layer: The velocity layer to use for evaluation.
    :type layer: :class:`~numpy.ndarray`, dtype = :py:const:`.VelocityLayer`
    :param prop: The material property to evaluate. One of:

        * ``p``
            Compressional (P) velocity (km/s)
        * ``s``
            Shear (S) velocity (km/s)
        * ``r`` or ``d``
            Density (g/cm^3)
    :type prop: str

    :returns: The value of the material property requested.
    :rtype: :class:`~numpy.ndarray` (dtype = :class:`float`, shape equivalent
        to ``layer``)
    """
    prop = prop.lower()
    if prop == "p":
        return layer['bot_p_velocity']
    elif prop == "s":
        return layer['bot_s_velocity']
    elif prop in "rd":
        return layer['bot_density']
    raise ValueError("Unknown material property, use p, s, or d.")


def evaluate_velocity_at_top(layer, prop):
    """
    Evaluate material properties at top of a velocity layer.

    .. seealso:: :func:`evaluate_velocity_at_bottom`,
        :func:`evaluate_velocity_at`

    :param layer: The velocity layer to use for evaluation.
    :type layer: :class:`~numpy.ndarray`, dtype = :py:const:`.VelocityLayer`
    :param prop: The material property to evaluate. One of:

        * ``p``
            Compressional (P) velocity (km/s)
        * ``s``
            Shear (S) velocity (km/s)
        * ``r`` or ``d``
            Density (g/cm^3)
    :type prop: str

    :returns: The value of the material property requested.
    :rtype: :class:`~numpy.ndarray` (dtype = :class:`float`, shape equivalent
        to ``layer``)
    """
    prop = prop.lower()
    if prop == "p":
        return layer['top_p_velocity']
    elif prop == "s":
        return layer['top_s_velocity']
    elif prop in "rd":
        return layer['top_density']
    raise ValueError("Unknown material property, use p, s, or d.")


def evaluate_velocity_at(layer, depth, prop):
    """
    Evaluate material properties at some depth in a velocity layer.

    .. seealso:: :func:`evaluate_velocity_at_top`,
        :func:`evaluate_velocity_at_bottom`

    :param layer: The velocity layer to use for evaluation.
    :type layer: :class:`~numpy.ndarray`, dtype = :py:const:`.VelocityLayer`
    :param depth: The depth at which the material property should be
        evaluated. Must be within the bounds of the layer or results will be
        undefined.
    :type depth: float
    :param prop: The material property to evaluate. One of:

        * ``p``
            Compressional (P) velocity (km/s)
        * ``s``
            Shear (S) velocity (km/s)
        * ``r`` or ``d``
            Density (g/cm^3)
    :type prop: str

    :returns: The value of the material property requested.
    :rtype: :class:`~numpy.ndarray` (dtype = :class:`float`, shape equivalent
        to ``layer``)
    """
    thick = layer['bot_depth'] - layer['top_depth']
    prop = prop.lower()
    if prop == "p":
        slope = (layer['bot_p_velocity'] - layer['top_p_velocity']) / thick
        return slope * (depth - layer['top_depth']) + layer['top_p_velocity']
    elif prop == "s":
        slope = (layer['bot_s_velocity'] - layer['top_s_velocity']) / thick
        return slope * (depth - layer['top_depth']) + layer['top_s_velocity']
    elif prop in "rd":
        slope = (layer['bot_density'] - layer['top_density']) / thick
        return slope * (depth - layer['top_depth']) + layer['top_density']
    raise ValueError("Unknown material property, use p, s, or d.")
